"""
Contains classes and utility/support functions for parsing DIII-D patch panel files
"""

try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import struct
import numpy as np
from omfit_classes.omfit_ascii import OMFITascii

__all__ = ['OMFITpatch', 'OMFITpatchObject', 'OMFITpatchList']


def lrl(item):
    """Shortcut to avoid writing list(range(len(...))) so many times. Does special format for long lists."""
    if len(item) > 5:
        return '0:{}'.format(len(item) - 1)
    return '{}'.format(list(range(len(item))))


class OMFITpatchList(list):
    """
    Handles list-like data within patch panels.

    __setitem__ notifies parent object of new assignment so values index/name pairs can be kept self consistent.
    Without this intervention, a list's element can be changed without changing the list itself and so the parent
    OMFITpatchObject instance gets no notification, and lists of chopper indices and chopper names can get out of sync.
    """

    def __init__(self, *args, **kw):
        self.OMFITproperties = {}
        parent = kw.pop('parent', None)
        assert isinstance(parent, (OMFITpatchObject, OMFITpatch)), 'parent object should be OMFITpatch or OMFITpatchObject'
        list.__init__(self, *args)
        self.parent = parent

    @property
    def parent(self):
        return self.OMFITproperties.get('parent', None)

    @parent.setter
    def parent(self, value):
        self.OMFITproperties['parent'] = value

    def __setitem__(self, idx, value):
        pkey = [k for k in self.parent if self.parent[k] is self][0]
        new_list = [v for v in self]
        new_list[idx] = value
        self.parent.__setitem__(pkey, new_list)

    def __copy__(self):
        """
        Instead of a true copy, return a mundane list.
        Special properties exist in relation to parent only, and that relationship can't be guaranteed during copy.
        """
        return [v for v in self]

    def __deepcopy__(self, memo):
        """
        Instead of a true copy, return a mundane list.
        Special properties exist in relation to parent only, and that relationship can't be guaranteed during copy.
        """
        return [copy.deepcopy(v) for v in self]


class OMFITpatchObject(SortedDict, OMFITobject):
    """
    Handles objects within a patch panel (such as a single F coil) and keeps indices & names consistent with each other
    """

    rail_map = ['E supply', 'VFI', '?', '']  # Maps rail_name to rail_index
    power_supply_map = {  # Maps power_supply_name to power_supply_index. Values < 1 are special & affect chopper, too.
        -3: '',  # RV2 or RV1
        0: '',  # SHORT
        1: 'D',
        2: 'V',
        3: 'T1',
        4: 'T2',
        5: 'HV1',
        6: 'HV2',
        8: 'D2',
    }
    # If the first chopper is 0, then it's named 'SHORT'. Otherwise, it's '' for unused.
    # If dealing with key 'cl' instead of an F-coil, it's 'OPEN'.
    chopper_map = (
        ['?']
        + ['X{}'.format(i + 1) for i in range(20)]
        + ['HX{}'.format(i + 1) for i in range(16)]
        + ['RV1', 'RV2']
        + ['?39', '?40', '?41', 'FSUP']
    )
    direction_map = ['', 'PUSH', 'PULL']

    allowed_choppers = {
        '1A': list(range(0, 21)) + [37, 38],
        '1B': list(range(0, 21)) + [37, 38],  # All X, RV1, RV2
        '2A': list(range(0, 21)) + [37, 38],
        '2B': list(range(0, 21)) + [37, 38],  # All X, RV1, RV2
        '3A': list(range(0, 21)) + [37, 38],
        '3B': list(range(0, 21)) + [37, 38],  # All X, RV1, RV2
        '4A': list(range(0, 21)) + [37, 38],
        '4B': list(range(0, 21)) + [37, 38],  # All X, RV1, RV2
        '5A': list(range(0, 21)) + [37, 38],
        '5B': list(range(0, 21)) + [37, 38],  # All X, RV1, RV2
        '6A': [0] + list(range(5, 13)) + list(range(21, 29)) + list(range(31, 37)),
        '6B': [0] + list(range(5, 13)) + list(range(21, 31)) + list(range(33, 37)),
        '7A': [0] + list(range(5, 13)) + list(range(21, 29)) + list(range(31, 37)),
        '7B': [0] + list(range(5, 13)) + list(range(21, 31)) + list(range(33, 37)),
        '8A': [0] + list(range(5, 13)) + [19, 20] + list(range(21, 27)) + list(range(31, 35)) + [37, 38],
        '8B': [0] + list(range(5, 13)) + [19, 20] + list(range(21, 27)) + list(range(33, 37)) + [37, 38],
        '9A': [0] + list(range(5, 13)) + [19, 20] + list(range(21, 27)) + list(range(31, 35)) + [37, 38],
        '9B': [0] + list(range(5, 13)) + [19, 20] + list(range(21, 27)) + list(range(33, 37)) + [37, 38],
        'cl': [0, 42],  # Leave it open or use FSUP to connect D-supply
        'HV1': list(range(21, 37)),
        'HV2': list(range(21, 37)),  # All HX choppers
        'D': list(range(1, 30)) + [42],  # All X, HX1-9 and FSUP to C-coils
        'V': list(range(1, 30)),  # All X, HX1-9
        'T1': list(range(1, 21)),  # All X choppers
        'T2': list(range(1, 21)) + list(range(21, 28)),  # All X, HX1-7
        'D2': list(range(1, 30)),  # All X choppers, HX1-9
    }
    cmc = {'{}{}'.format(i, a): 3 for i in [1, 2, 3, 4, 5, 8] for a in 'AB'}
    cmc['9A'] = cmc['9B'] = 4
    cmc['6A'] = cmc['6B'] = cmc['7A'] = cmc['7B'] = 6
    cmc['cl'] = 1
    chopper_max_counts = cmc

    lonely_choppers = [37, 38, 42]  # These chopper-like items aren't allowed to be in parallel with anything

    def __init__(self, filename=None, **kw):
        self.OMFITproperties = {}
        self.locked = False
        OMFITobject.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.filename = filename  # This doesn't actually have a filename; it just can't be missing.
        return

    @property
    def locked(self):
        return self.OMFITproperties.get('locked', None)

    @locked.setter
    def locked(self, value):
        self.OMFITproperties['locked'] = value
        return

    def __copy__(self):  # https://stackoverflow.com/a/48339837/6605826
        """
        Special copy method that unlocks, does copy, then restores locked status.
        Without the unlock first, there can be an infinite loop as the special __setitem__ fights itself.
        """
        result = OMFITpatchObject()
        result.locked = False
        for k, v in self.items():
            result[k] = self[k]
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):  # https://stackoverflow.com/a/15774013/6605826
        """
        Special copy method that unlocks, does copy, then restores locked status.
        Without the unlock first, there can be an infinite loop as the special __setitem__ fights itself.
        """
        result = OMFITpatchObject()
        memo[id(self)] = result
        result.locked = False
        for k, v in self.items():
            result[copy.deepcopy(k)] = copy.deepcopy(self[k])
            if isinstance(self[k], OMFITpatchList):
                result[k] = OMFITpatchList(result[k], parent=result)

        for k, v in self.__dict__.items():
            if k in ['parent', '_OMFITparent']:
                setattr(result, k, None)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def __setitem__(self, key, value):
        """
        The point of this special setitem method is to keep names and indices in sync.
        For example, changing 'rail_name' will also change 'rail_index'.
        If you want to mess with the file without synchronized updates taking place,
        unlock by setting self.locked = False and then this method will pass through
        to the super class's setitem method.
        :param key:
            Key for looking up stuff in self
        :param value:
            Value to assign to self[key]
        :return: None
        """
        if not getattr(self, 'locked', False):
            # This instance is unlocked, so don't do synchronized updates.
            SortedDict.__setitem__(self, key, value)
            return

        # Handle power supply rails
        if key == 'rail_name':
            other_key = 'rail_index'
            if value not in self.rail_map:
                raise ValueError('Invalid power rail name: {}. Valid = {}'.format(value, repr(self.rail_map)))
            if value == '':
                other_value = 3  # This is for D supply on C coils, which is a special case.
                # There may be other weird setups
            else:
                other_value = self.rail_map.index(value)
        elif key == 'rail_index':
            value = int(value)
            other_key = 'rail_name'
            if (value >= len(self.rail_map)) or (value < 0):
                raise ValueError('Invalid power rail index: {}. Valid = {}'.format(value, lrl(self.rail_map)))
            other_value = self.rail_map[value]

        # Handle current directions
        elif key == 'direction_name':
            other_key = 'direction_index'
            if value not in self.direction_map:
                raise ValueError('Invalid current direction name: {}. Valid = {}'.format(repr(value), repr(self.direction_map)))
            other_value = self.direction_map.index(value)
        elif key == 'direction_index':
            value = int(value)
            other_key = 'direction_name'
            if (value >= len(self.direction_map)) or (value < 0):
                raise ValueError('Invalid current dir index: {}. Valid = {}'.format(value, lrl(self.direction_map)))
            other_value = self.direction_map[value]

        # Handle choppers
        elif key == 'chopper_name':
            other_key = 'chopper_index'
            for val in value:
                if val not in (self.chopper_map + ['', 'SHORT', 'OPEN', 'SHOR']):
                    raise ValueError('Invalid chopper name: {}. Valid = {}'.format(repr(val), repr(self.chopper_map)))
            other_value = [self.chopper_map.index(val) if val in self.chopper_map else 0 for val in value]
            new_count = np.sum([ov != 0 for ov in other_value])
            SortedDict.__setitem__(self, 'chopper_count', new_count)
            if ('power_supply_name' in self) and (self['power_supply_name'] == ''):
                new_power_supply_index = 0 if other_value[0] == 0 else -3
                SortedDict.__setitem__(self, 'power_supply_index', new_power_supply_index)
        elif key == 'chopper_index':
            value = [int(val) for val in value]
            other_key = 'chopper_name'
            for val in value:
                if (val >= len(self.chopper_map)) or (val < 0):
                    raise ValueError('Invalid chopper index: {}. Valid = {}'.format(val, lrl(self.chopper_map)))
            other_value = [self.chopper_map[val] for val in value]
            # Resolve stuff stored in the 0 index
            if other_value[0] == '?':
                if self['coil_index'] == 19:
                    other_value[0] = 'OPEN'
                else:
                    other_value[0] = 'SHORT'
            for i in range(1, len(other_value)):
                if other_value[i] == '?':
                    other_value[i] = ''
            new_count = np.sum([ov != '' for ov in other_value])
            SortedDict.__setitem__(self, 'chopper_count', new_count)
            if self['power_supply_name'] == '':
                new_power_supply_index = 0 if value[0] == 0 else -3
                SortedDict.__setitem__(self, 'power_supply_index', new_power_supply_index)
        elif key == 'chopper_count':
            raise ValueError('Do not change chopper count manually')

        # Handle power supplies
        elif key == 'power_supply_name':
            other_key = 'power_supply_index'
            if value not in list(self.power_supply_map.values()):
                raise ValueError(
                    'Invalid power supply name: {}. Valid = {}'.format(repr(value), repr(list(self.power_supply_map.values())))
                )
            other_value = [k for k, v in self.power_supply_map.items() if v == value]
            # Resolve possible duplicates that have '' as the value
            if (self['chopper_index'][0] > 36) and (0 in other_value):
                other_value = [ov for ov in other_value if ov < 0]
            elif (self['chopper_index'][0] == 0) and (min(other_value) < 0):
                other_value = [ov for ov in other_value if ov >= 0]
            if len(other_value) == 1:
                other_value = other_value[0]
            else:
                raise ValueError('Values in power supply map should be unique after handling values of ""')
        elif key == 'power_supply_index':
            other_key = 'power_supply_name'
            if value not in self.power_supply_map:
                raise ValueError('Invalid power supply index: {}. Valid = {}'.format(value, repr(list(self.power_supply_map.keys()))))
            other_value = self.power_supply_map[value]

        # Contingency
        else:
            other_key = other_value = None

        # The OMFITpatchList class notifies its parent when an element is updated, so replace lists with that.
        if isinstance(value, list):
            value = OMFITpatchList(value, parent=self)
        if isinstance(other_value, list):
            other_value = OMFITpatchList(other_value, parent=self)

        # Write the new values to self
        SortedDict.__setitem__(self, key, value)
        if other_key is not None and other_value is not None:
            SortedDict.__setitem__(self, other_key, other_value)

        # Update status
        if (getattr(self, '_OMFITparent', None) is not None) and self.locked:
            self._OMFITparent.check()

        return None

    def __tree_repr__(self):
        """Provides custom tree representation for this object"""
        try:
            patch_type = self._OMFITparent.patch_type
        except AttributeError:
            patch_type = '?'

        if patch_type in ['F', 'P']:
            chopper_note = self['chopper_name'][0] if self['chopper_count'] == 1 else '{} choppers'.format(self['chopper_count'])
            tree_repr = '{rail:10s} {ps:8s} {dir:8s} {chopper_note:}'.format(
                rail=self['rail_name'], ps=self['power_supply_name'], dir=self['direction_name'], chopper_note=chopper_note
            )
        elif patch_type == 'I':
            tree_repr = []
            if 'c_coil_parity' in self:
                tree_repr += ['C parity: {}'.format(self['c_coil_parity'])]
            if 'i_coil_file' in self:
                tree_repr += ['I-coils: {}'.format(self['i_coil_file'])]
            if np.all([a in self for a in ['slot', 'sign', 'spa', 'c_idx', 'coil']]):
                tree_repr += ['{slot:2s} {sign:3s} {spa:4s} {c_idx:1d} {coil:}'.format(**self)]
            if np.all(['AA{}'.format(i) in self for i in range(1, 13)]):
                aa = []
                for i in range(1, 13):
                    if self['AA{}'.format(i)]:
                        aa += ['AA{}:{}'.format(i, self['AA{}'.format(i)])]
                tree_repr += aa if aa else ['No AA used']

            tree_repr = ', '.join(tree_repr)
        else:
            tree_repr = None
        return tree_repr, []


# List attributes which need to be saved with projects.
OMFITpatch_save_attrs = [
    'debug_topic',
    'patch_type',
    'auto_clean',
    'status',
    'contents',
    'original_contents',
    'patch_name',
    'original_patch_name',
    'original_patch_type',
    'fpconvert',
    'shot',
    '_load_complete',
    'server',
    'tunnel',
    'remote_dir',
    'work_dir',
    # Deliberately not listed:
    # 'problems',
]


class OMFITpatch(SortedDict, OMFITascii):
    """
    Parses DIII-D PCS patch panel files.

    Several types of files are recognized:
    - Type-F: F-coil patch panel in ASCII archival format.
    - Type-P: F-coil patch panel in binary format. Can be converted to type-F by
      changing .patch_type attribute and executing .add_c_coil() method. May be obsolete.
    - Type-I: I&C coil patch panel in ASCII format.

    """

    # Define the marker for P-files with binary data. Also split them here.
    bin_mark = r'\x00'

    def __getattr__(self, attr_name):
        """
        Custom attribute access method.
        There are too many funky attributes and I don't want to do a @property
        declaration for each one, so we're doing this instead.
         - https://docs.python.org/3/reference/datamodel.html#customizing-attribute-access
         - https://stackoverflow.com/a/52729406/6605826

        :param attr_name: string
            Name of the attribute

        :return: Attribute value
        """
        if attr_name in OMFITpatch_save_attrs:
            return self.OMFITproperties.get(attr_name, None)
        return SortedDict.__getattr__(self, attr_name)

    def __setattr__(self, attr_name, value):
        """
        Allows handling of a large number of attributes without @property declarations
        :param attr_name: string
        :param value:
        :return: None
        """
        if attr_name in OMFITpatch_save_attrs:
            self.OMFITproperties[attr_name] = value
            if attr_name.startswith('custom_'):
                # Setting a custom attribute should update the attribute it overrides, as it does in init
                setattr(self, attr_name[len('custom_') :], value)
            return
        SortedDict.__setattr__(self, attr_name, value)
        return

    def __delattr__(self, attr_name):
        """
        Allows handling of a large number of attributes without @property declarations
        :param attr_name: string
        :return: None
        """
        if attr_name in OMFITpatch_save_attrs:
            self.OMFITproperties[attr_name] = None
            if attr_name.startswith('custom_'):
                # Deleting a custom attribute should revert its counterpart to default
                setattr(self, attr_name[len('custom_') :], getattr(self, 'default_' + attr_name[len('custom_') :], None))
            return
        SortedDict.__delattr__(self, attr_name)
        return

    def __init__(
        self,
        filename=None,
        shot=None,
        patch_type=None,
        debug_topic='OMFITpatch',
        auto_clean=True,
        fpconvert=True,
        server=None,
        tunnel=None,
        work_dir=None,
        remote_dir=None,
        default_patch_type='F',
        **kw,
    ):
        """
        :param filename: string [optional if shot is provided]
            Filename of original source file, including path. This will be preserved as
            self.source_file, even if the class updates to a temporary file copy.
            If shot is provided, filename controls output only; no source file is read.
            In this case, filename need not include path.

        :param shot: int [optional if filename is provided]
            Shot number to use to look up patch data. Must provide patch_type when using shot.
            If a filename is provided as well, shot will be used for lookup and filename will control output only.

        :param patch_type: None or string
            None lets the class auto-assign, which should be fine.
            You can force it explicitly if you really want to.

        :param debug_topic: string
            Topic keyword to pass to printd. Allows all printd from the class to be consistent.

        :param auto_clean: bool
            Run cleanup method after parsing. This will remove some problems,
            but prevent exact recovery of problematic original contents.

        :param fpconvert: bool
            Automatically convert type P files into type F so they can be saved.

        :param server: string [optional if running in OMFIT framework]
            Complete access instruction for server that runs viwhed, like "eldond@omega.gat.com:22"

        :param tunnel: string [optional if running in OMFIT framework]
            Complete access instruction for tunnel used to reach server, like "eldond@cybele.gat.com:2039".
            Use empty string '' if no tunnel is needed.

        :param work_dir: string  [optional if running in OMFIT framework]
            Local working directory (temporary/scratch) for temporary files related to remote executable calls

        :param remote_dir: string  [optional if running in OMFIT framework]
            Remote working directory (temporary/scratch) for temporary files related to remote executable calls

        :param default_patch_type: string
            Patch panel type to assign if auto detection fails, such as if you're initializing a blank patch panel file.
            If you are reading a valid patch panel file, auto detection will probably work because it is very good.
            Please choose from 'F' or 'I'. 'P' is a valid patch_type, but it's read-only so it's a very bad choice for
            initializing a blank file to fill in yourself.

        :param kw: Other keywords passed to OMFITascii
        """
        self.OMFITproperties = {}
        self.source_file = copy.copy(filename) if shot is None else shot
        SortedDict.__init__(self)
        if filename is None:
            fn = funny_random_name_generator(patch_type == 'I')
        elif filename.endswith(os.sep):
            fn = filename + funny_random_name_generator(patch_type == 'I')
        else:
            fn = filename
        if os.sep in fn:
            from pathlib import Path

            Path(fn).touch()  # OMFITobjects don't like getting a filename with a path unless the file exists
        OMFITascii.__init__(self, fn, **kw)
        self._load_complete = kw.pop('_load_complete', False)
        self.shot = shot
        self.server = server
        self.tunnel = tunnel
        self.remote_dir = remote_dir
        self.work_dir = work_dir
        self.debug_topic = debug_topic
        self.status = 'init'
        self.auto_clean = auto_clean
        self.fpconvert = fpconvert
        self.patch_type = patch_type
        if ((shot is None) or self._load_complete) and (filename is not None):
            try:
                with open(self.filename, 'r') as f:
                    self.contents = f.read()
            except UnicodeDecodeError:
                with open(self.filename, 'rb') as f:
                    self.contents = str(f.read())
        elif shot is not None:
            self.contents = self._read_shot(default_patch_type)
        else:
            self.contents = ''
        self.original_contents = copy.copy(self.contents)
        self.patch_type = self._guess_patch_type(default=default_patch_type) if patch_type is None else patch_type
        self.original_patch_type = copy.copy(self.patch_type)
        self.patch_name = getattr(self, 'patch_name', None)  # Allows for possibility that it was set by OMFITproperties
        if self.patch_type == 'F':
            if (self.shot is None) or self._load_complete:
                # The type-F method is suitable from a type-F FILE or from contents from a shot after parsing
                self._read_type_f()
            else:
                # The type-P method is suitable for contents read directly from a shot
                self._read_type_p_inner(self.contents)
                if self.auto_clean:
                    self.printq('Running automatic cleanup of F-patch file to remove junk chopper data')
                    self._clean_type_f()
        elif self.patch_type == 'P':
            self._read_type_p()
        elif self.patch_type == 'I':
            if (self.shot is None) or self._load_complete:
                self._read_type_i()
            else:
                self._read_type_j()
        else:
            raise ValueError('Invalid or unhandled patch panel file.')
        self.check()
        self.original_patch_name = copy.copy(self.patch_name)
        self.problems = []

        if self.filename is None and self.patch_name is not None:
            # This must've been read in from a shot numbers instead of a file. Let's make a real filename.
            # Generate a temporary directory name
            subprocess_dir = '_'.join(map(str, OMFITaux['prun_process']))
            if len(subprocess_dir):
                subprocess_dir = '__p' + subprocess_dir
            directory = (
                OMFITcwd + os.sep + 'objects' + os.sep + 'file_' + utils_base.now("%Y-%m-%d__%H_%M" + subprocess_dir + os.sep + "%S__%f")
            )
            while os.path.exists(directory):
                directory += "_"
            # Assign the new filename
            self.filename = directory + os.sep + self.patch_name
            if self.patch_type in ['I', 'F']:
                if not os.path.exists(self.filename):
                    os.makedirs(os.path.split(self.filename)[0])
                self.save()
            self.printq('Auto-assigned filename {}'.format(self.filename))

        # Make sure all saved attributes are restored by allowing them to be passed in through **kw
        for attr in OMFITpatch_save_attrs:
            kw_attr = kw.pop(attr, None)
            if kw_attr is not None:
                setattr(self, attr, kw_attr)

        if ((filename is None) or filename.endswith(os.sep)) and self.patch_name is not None:
            # Replace "working title" filename with good patch name if we were able to find one
            self.filename = self.patch_name if filename is None else filename + self.patch_name
        self._load_complete = True  # Prevents instance from saved project from re-doing read from shot or whatever
        return

    def __tree_repr__(self):
        """Forms strings to display in OMFIT tree GUI to represent instances of this class"""
        file_rep = OMFITascii.__tree_repr__(self)[0]
        tree_rep = file_rep + ', STATUS: {}, PATCH TYPE: {}, NAME: {}'.format(self.status, self.patch_type, self.patch_name)
        return tree_rep, []

    def _read_shot(self, default_patch_type):
        """Reads patch data from a shot number"""

        if (self.patch_type is None) and (default_patch_type in ['F', 'I', 'P']):
            self.patch_type = default_patch_type
            self.original_patch_type = default_patch_type
            self.printq('Assigned patch_type to default ("{}") while reading from shot.'.format(default_patch_type))

        if self.patch_type in ['F', 'P']:
            patch_code = 'PATCH'
        elif self.patch_type == 'I':
            patch_code = 'SPATCH'
        else:
            raise ValueError(
                'Bad patch_type={}. Please specify patch_type="I" or patch_type="F". True automatic patch_type '
                'determination is not possible when reading from a shot number.'.format(repr(self.patch_type))
            )

        executable = 'module load defaults\nviwhed'
        interactive = ['sh {}'.format(self.shot), 'na {}'.format(patch_code), 'df', 'ex']
        stdout_catch = []
        stderr_catch = []
        self._set_default_server_info()
        import omfit_classes.OMFITx as OMFITx

        OMFITx.executable(
            None,
            [],
            [],
            server=self.server,
            tunnel=self.tunnel,
            remotedir=self.remote_dir,
            workdir=self.work_dir,
            std_out=stdout_catch,
            std_err=stderr_catch,
            executable=executable,
            interactive_input='\n'.join(interactive),
        )
        i = 0
        props = 'server = {server:}, tunnel = {tunnel:}, remote_dir = {remote_dir:}, work_dir = {work_dir:}'.format(**self.OMFITproperties)

        while (i < len(stdout_catch)) and (' ASCII -' not in stdout_catch[i]):
            i += 1
        if i == len(stdout_catch):
            printe('ERROR in OMFITpatch._read_shot: Reached end of stdout without finding expected ASCII marker!')
            print('-' * 79)
            print('\n'.join(stdout_catch))
            print('-' * 79)
            print(props)
            raise ValueError('Reached end of stdout without finding expected ASCII marker!')
        j = i + 1
        while (j < len(stdout_catch)) and (' INT16 - ' not in stdout_catch[j]):
            j += 1
        if i == len(stdout_catch):
            printe('ERROR in OMFITpatch._read_shot: Reached end of stdout without finding expected INT16 marker!')
            print('-' * 79)
            print('\n'.join(stdout_catch))
            print('-' * 79)
            print(props)
            raise ValueError('Reached end of stdout without finding expected INT16 marker!')

        return ''.join([line.split('"')[1] for line in stdout_catch[i + 1 : j]])

    def _set_default_server_info(self):
        """If any of server, tunnel, work_dir, or remote_dir are None, update them with default values"""
        import getpass

        server_name = 'omega'  # The server options are VERY limited. It's okay to leave this at just omega.
        if None in [self.server, self.tunnel, self.remote_dir, self.work_dir]:
            # Need to load defaults
            dir_suffix = os.sep + str(datetime.datetime.now()).replace(' ', '_').replace(':', '_')
            # noinspection PyBroadException
            try:
                default_server = evalExpr(OMFIT['MainSettings']['SERVER'][server_name]['server'])
                default_remote_dir = evalExpr(OMFIT['MainSettings']['SERVER'][server_name]['workDir']) + dir_suffix
                localhost = evalExpr(OMFIT['MainSettings']['SERVER']['localhost'])
                if isinstance(localhost, str):
                    localhost = evalExpr(OMFIT['MainSettings']['SERVER'][localhost])
                default_work_dir = evalExpr(localhost['workDir']) + dir_suffix
                default_tunnel = evalExpr(OMFIT['MainSettings']['SERVER'][server_name]['server'])
            except Exception:
                # This plan used to be relevant, but maybe not anymore?
                # It might trigger on KeyError for 'MainSettings' not being in OMFIT, if that can ever happen
                self.printq(
                    'Failed to reference OMFIT MainSettings. Guessing default settings for remote server access. '
                    'If these do not work, try specifying server, tunnel, remote_dir, and work_dir next time.'
                )
                username = getpass.getuser()
                default_server = '{}@omega.gat.com:22'.format(username)
                default_tunnel = '{}@cybele.gat.com:2039'.format(username)
                default_work_dir = tempfile.mkdtemp('_OMFIT_patch_read_shot') + os.sep
                default_remote_dir = os.sep.join(['cluster-scratch', username, 'OMFIT', dir_suffix])
            self.server = default_server if self.server is None else self.server
            self.tunnel = default_tunnel if self.tunnel is None else self.tunnel
            self.remote_dir = default_remote_dir if self.remote_dir is None else self.remote_dir
            self.work_dir = default_work_dir if self.work_dir is None else self.work_dir
        else:
            self.printq('server, work_dir, and remote_dir settings are already good; no need to load defaults')

    def check(self):
        """Checks for problems and returns a status string that is either 'OKAY' or a list of problems."""
        if self.patch_type in ['F', 'P']:
            status = self._check_type_f()
        else:
            status = '?'
        self.status = status
        return status

    def _check_type_f(self):
        """
        Specific checking method for type-F or P patch panel files.
        :return: string
            "OKAY" if no problems are found. Otherwise, a terse description of problems.
        """
        assert self.patch_type in ['F', 'P'], 'This method only applies to patch_type F or P, not {}'.format(self.patch_type)

        problems = []
        coils = [k for k in self if isinstance(self[k], OMFITpatchObject)]

        # Check for duplicate choppers
        all_chopper_name = []
        for k in coils:
            all_chopper_name += self[k]['chopper_name']
        ignores = ['', 'SHORT', 'OPEN']
        for ignore in ignores:
            while ignore in all_chopper_name:
                all_chopper_name.remove(ignore)
        duplicate_choppers = []
        noted = []
        for chop in all_chopper_name:
            chop_count = all_chopper_name.count(chop)
            if chop_count > 1:
                if chop not in noted:
                    c_on_coils = [k for k in coils if chop in self[k]['chopper_name']]
                    duplicate_choppers += ['{}({})'.format(chop, ','.join(c_on_coils))]
                    noted += [chop]
        if len(duplicate_choppers):
            problems += ['DUPL.CHOP: [' + ', '.join(duplicate_choppers) + ']']

        # Check for mis-matched current directions for the same power supply
        ps_map = OMFITpatchObject.power_supply_map
        pwr_dirs = {k: None for k in ps_map if k > 0}
        coils_used = {k: [] for k in ps_map if k > 0}
        for k in coils:
            ps = self[k]['power_supply_index']
            if ps in pwr_dirs:
                coils_used[ps] += [k]
                if pwr_dirs[ps] is None:
                    pwr_dirs[ps] = self[k]['direction_index']
                else:
                    if pwr_dirs[ps] != self[k]['direction_index']:
                        pwr_dirs[ps] = -1
        dir_problems = []
        for k in pwr_dirs:
            if pwr_dirs[k] == -1:
                dir_problems += ['{}({})'.format(ps_map[k], ','.join(coils_used[k]))]
        if len(dir_problems):
            problems += ['DIR.CONFL.: [' + ', '.join(dir_problems) + ']']

        # Check for missing C-coil
        if 'cl' not in self:
            problems += ['MISSING.C.CL']

        if len(self) > 0:
            # Can't have these problems if there are no coils defined

            # Check for too many choppers
            cmc = self[coils[0]].chopper_max_counts
            too_many_choppers = []
            for coil in coils:
                if coil not in cmc:
                    too_many_choppers += ['{}({}>?)'.format(coil, self[coil]['chopper_count'])]
                elif self[coil]['chopper_count'] > cmc[coil]:
                    too_many_choppers += ['{}({}>{})'.format(coil, self[coil]['chopper_count'], cmc[coil])]
            if too_many_choppers:
                problems += ['EXCESS.CHOP.: [' + ', '.join(too_many_choppers) + ']']

            # Check for illegal chopper vs. coil and power supply pairings
            ac = self[coils[0]].allowed_choppers
            illegal_choppers = []
            for coil in coils:
                illp = []
                illc = []
                psn = self[coil].get('power_supply_name', '')
                if coil not in ac:
                    self.printq('Coil {} is not recognized in dict of allowed choppers; cannot check properly'.format(coil))
                for ci in self[coil]['chopper_index']:
                    if (coil in ac) and (ci not in ac[coil]):
                        illc += [self[coil].chopper_map[ci]]
                    if (psn in ac) and (ci != 0) and (ci not in ac[psn]):
                        illp += ['{}:{}'.format(self[coil].chopper_map[ci], coil)]
                if illc:
                    illegal_choppers += ['{}({})'.format(coil, ','.join(illc))]
                if illp:
                    illegal_choppers += ['{}({})'.format(psn, ','.join(illp))]
            if illegal_choppers:
                problems += ['ILLEG.CHOP.: [' + ', '.join(illegal_choppers) + ']']

            bad_parallel_choppers = []
            for coil in coils:
                # Check for parallel chopper-like objects that aren't allowed to be in parallel
                for lc in self[coils[0]].lonely_choppers:
                    if (lc in self[coil]['chopper_index']) and (self[coil]['chopper_count'] > 1):
                        names = [cn for cn in self[coil]['chopper_name'] if cn]
                        bad_parallel_choppers += ['{}({})'.format(coil, ', '.join(names))]
            if bad_parallel_choppers:
                problems += ['BAD.PARALLEL.: [' + ', '.join(list(set(bad_parallel_choppers))) + ']']
        else:
            problems += ['NO.COILS.']

        # Summarize problems
        if problems:
            status = ', '.join(problems)
        else:
            status = 'OKAY'
        self.problems = problems
        return status

    def printq(self, *args):
        """Print wrapper for keeping topic consistent within the instance"""
        printd(*args, topic=self.debug_topic)

    def _guess_patch_type(self, default='?'):
        """
        Differentiates between various file types and returns a string or None.
        :return:
            'P': PCS format, like ptpatch.dat. Similar to 'F' type, but includes strange binary data
            'F': F-coil patch panel, ASCII
            'I': I&C-coil patch panel, ASCII
            None: Failed to determine type
        """
        if self.bin_mark in self.contents:
            return 'P'
        elif 'C-coil parity:' in self.contents:
            return 'I'
        elif 'I-coil file:' in self.contents:
            return 'I'
        elif (('PUSH' in self.contents) or ('PULL' in self.contents)) and ('1A' in self.contents):
            return 'F'
        else:
            return default

    def _read_type_f(self):
        """Parses F-coil patch panel files into dict-like structure"""
        assert self.patch_type == 'F', 'This method only works for type "F". Type is instead: %s' % self.patch_type
        field_widths = [3, 4, 3, 6, 3, 4, 2, 5, 2, 2, 5, 5, 5, 5, 5, 3, 3, 3, 3, 3]

        for i, line in enumerate(self.contents.split('\n')):
            if len(line) != np.sum(field_widths):
                self.printq('Line {} does not match expected length ({}); contents = {}'.format(i, np.sum(field_widths), line))
                continue
            fields = [''] * len(field_widths)
            start = 0
            for j in range(len(field_widths)):
                fields[j] = line[start : start + field_widths[j]]
                start += field_widths[j]
            out = self[fields[1].strip()] = OMFITpatchObject()
            out['coil_index'] = int(fields[0])
            out['chopper_count'] = int(fields[9])
            out['chopper_index'] = [int(fields[2])]
            out['power_supply_index'] = int(fields[4])
            out['direction_index'] = int(fields[6])
            out['rail_index'] = int(fields[8])
            out['chopper_name'] = [fields[3].strip()]
            out['power_supply_name'] = fields[5].strip()
            out['direction_name'] = fields[7].strip()
            if out['rail_index'] == 0:
                out['rail_name'] = 'E supply'
            elif out['rail_index'] == 1:
                out['rail_name'] = 'VFI'
            else:
                out['rail_name'] = ''
            out['chopper_name'] = [out['chopper_name'][0]] + [cn.strip() for cn in fields[10:15]]
            out['chopper_index'] = [out['chopper_index'][0]] + [int(ci) for ci in fields[15:20]]
            out['chopper_name'] = OMFITpatchList(out['chopper_name'], parent=out)
            out['chopper_index'] = OMFITpatchList(out['chopper_index'], parent=out)
            out.locked = True

        self.get_patch_name()

        if self.auto_clean:
            self.printq('Running automatic cleanup of F-patch file (archive format/type F) to remove junk chopper data')
            self._clean_type_f()

    def _read_type_p(self, include_c_coil=False):
        """Parses type-P patch panel files, which are F-patch files in binary format"""
        self._read_type_p_inner(self.contents.split(self.bin_mark)[45])

        if include_c_coil:
            self.add_c_coil()
            # This might be needed to make converted type-P to F files consistent with natural type-F files

        if self.auto_clean:
            self.printq('Running automatic cleanup of F-patch file (binary format/type P) to remove junk chopper data')
            self._clean_type_f()

        if self.fpconvert:
            printd('Automatically converting from type-P (binary) to type-F (ASCII) to enable saving.')
            self.patch_type = 'F'
            if 'cl' not in self:
                printw('WARNING: please use .add_c_coil(d_supply=True/False) to add C-coil to converted P-type patch!')

    def _read_type_p_inner(self, item_names):
        """Splits apart a string that's formatted like the relevant section of a type-P file"""
        fcoil_len = 36  # Number of characters for each F-coil
        title_len = 12  # Number of characters for title
        self.patch_name = item_names[:title_len].strip()
        rest = item_names[title_len:]
        coil_index = 1
        while len(rest) >= fcoil_len:
            out = self[rest[:4].strip()] = OMFITpatchObject()
            out['coil_index'] = coil_index
            # Load placeholders so the order will be consistent with F-type
            out['chopper_count'] = 0
            out['chopper_index'] = [0, 0, 0, 0, 0, 0]
            out['power_supply_index'] = 0
            out['direction_index'] = 0
            out['rail_index'] = 0
            # Done with placeholders
            out.locked = True
            out['chopper_name'] = [rest[16 + i * 4 : 20 + i * 4].strip() for i in range(5)] + ['']  # P-type only holds 5 choppers
            ps = rest[4:8].strip()
            out['power_supply_name'] = '' if ps == 'NONE' else ps
            out['direction_name'] = rest[12:16].strip()
            out['rail_name'] = 'VFI' if rest[8:11] == 'YES' else 'E supply'
            rest = rest[fcoil_len:]
            coil_index = coil_index + 1
        assert len(rest) == 0, 'Inner P contents after read FC should be empty, but instead: len{} remainder={}'.format(
            len(rest), repr(rest)
        )
        return

    def _read_type_i(self):
        """Parses I & C coil files"""
        assert self.patch_type == 'I', 'This method only works for type "I". Type is instead: %s' % self.patch_type
        if self.contents:
            lines = self.contents.split('\n')
            self._read_type_i_inner(lines)
        self.get_patch_name()

    def _read_type_i_inner(self, lines, xs=0):
        self['config'] = OMFITpatchObject()
        self['config']['c_coil_parity'] = lines[0][15:].strip()
        self['config']['i_coil_file'] = lines[11][13:].strip()
        for i in range(10):
            line = lines[1 + i]
            out = self['c{}'.format(i)] = OMFITpatchObject()
            out['slot'] = line[2:4]
            out['sign'] = line[5:8]
            out['spa'] = line[9:13].strip()
            out['c_idx'] = int(line[14 + xs : 15 + xs])
            out['coil'] = line[16:].strip()
        self['AA'] = OMFITpatchObject()
        for i in range(12):
            line = lines[12 + i]
            self['AA'][line[:5].strip()] = line[5:].strip()
        self['SS'] = OMFITpatchObject()
        if len(lines) > 25:
            # Has SS entries
            for i in range(12):
                line = lines[24 + i]
                self['SS'][line[:7].strip()] = line[7:].strip()
        else:
            # Old file with no SS entries, set them all disconnected
            for i in [1, 2]:
                for a in 'ABCDEF':
                    self['SS']['SS{}{}'.format(i, a)] = ''

    def _read_type_j(self):
        """Parses I & C patch information obtained from a shot number"""
        line_len1 = 36
        line_len2 = 10
        line_lens = [line_len1] * 10 + [32, 4] + [line_len2] * 12 + [line_len1] * 12
        rest = copy.copy(self.contents)
        lines = []
        for line_len in line_lens:
            lines += [rest[:line_len]]
            rest = rest[line_len:]
        self.patch_name = rest.strip()
        lines_out = ['C-coil parity: ' + lines[11]] + lines[0:10] + ['I-Coil file: ' + lines[10]] + lines[12:]
        lines_out = [line.replace(' pos', 'pos').replace(' neg', 'neg') for line in lines_out]
        self._read_type_i_inner(lines_out, xs=1)

    def get_patch_name(self):
        """Tries to determine patch panel name"""
        if (getattr(self, '_load_complete', None) is not None) and (getattr(self, 'patch_name', None) is not None):
            # Already have patch_name determined by previous load
            return
        elif (self.source_file is not None) and os.path.split(self.source_file)[-1].lower().endswith('patnow.dat'):
            name_file = self.source_file.replace('patnow.dat', 'patnow.nam')
            if os.path.exists(name_file):
                with open(name_file, 'r') as f:
                    patch_dat = f.read()
                    self.patch_name = patch_dat.split('\n')[0]
                return
        elif getattr(self, 'filename', None) is not None:
            self.patch_name = os.path.split(self.filename)[-1]
            return

        self.patch_name = funny_random_name_generator(self.patch_type == 'I')
        return

    def add_c_coil(self, d_supply=False):
        """
        Adds an entry for C-coils to type-F patch panels, which is useful if converted from type-P.
        Type-P files don't have a C-coil entry.
        :param d_supply: bool
            True: Add the default D supply on C-coil setup.
            False: Add default C-coil setup with no F-coil supplies on the C-coils (C-coil entry in F is blank)
        """
        if 'cl' in self:
            self.printq('This patch panel already has C-coils. Skipping.')
            return

        out = self['cl'] = OMFITpatchObject()
        out['coil_index'] = max([a.get('coil_index', 0) for a in self.values() if isinstance(a, OMFITpatchObject)]) + 1
        if d_supply:
            out['chopper_count'] = 1
            out['chopper_index'] = [42, 0, 0, 0, 0, 0]
            out['power_supply_index'] = 1
            out['direction_index'] = 1
            out['rail_index'] = 3
            out['chopper_name'] = ['FSUP', '', '', '', '', '']
            out['power_supply_name'] = 'D'
            out['direction_name'] = 'PUSH'
            out['rail_name'] = ''
        else:
            out['chopper_count'] = 1
            out['chopper_index'] = [0, 0, 0, 0, 0, 0]
            out['power_supply_index'] = 0
            out['direction_index'] = 0
            out['rail_index'] = 1
            out['chopper_name'] = ['OPEN', '', '', '', '', '']
            out['power_supply_name'] = ''
            out['direction_name'] = ''
            out['rail_name'] = 'VFI'
        out.locked = True

    def _clean_type_f(self):
        """Removes unused chopper indices and other problems"""
        assert self.patch_type in ['F', 'P'], 'This method should only be used for type F or P patch panels'
        coils = [k for k in self if isinstance(self[k], OMFITpatchObject)]
        for coil in coils:
            a = self[coil]
            # If the first chopper isn't a real chopper, any other indices specified are junk
            if (a['chopper_index'][0] < 1) or (a['chopper_name'][0] in ['RV1', 'RV2', 'OPEN', 'SHORT', '']):
                for i in range(1, 6):
                    a['chopper_index'][i] = 0

    def cleanup(self):
        """Cleans up junk data and blatant inconsistencies. May prevent output from matching buggy input."""
        if self.patch_type in ['F', 'P']:
            self._clean_type_f()
        else:
            print('No cleanup method for patch panel type {}'.format(self.patch_type))

    def _write_type_f(self):
        """Method for forming output string (file contents) for type-F patch panel files."""
        out_format = (
            '{coil_index:3d} {coil:>3s}{c0i:3d} {c0n:5s}{power_supply_index:3d} {power_supply_name:3s}'
            '{direction_index:2d} {direction_name:4s}{rail_index:2d}{chopper_count:2d} {c1n:24s}{c1i:15s}'
        )
        output = []
        for i, coil in enumerate([k for k in self if 'coil_index' in self[k]]):
            d = dict(**self[coil])
            d['c0i'] = d['chopper_index'][0]
            d['c0n'] = d['chopper_name'][0]
            d['c1n'] = ' '.join(['{:4s}'.format(cn) for cn in d['chopper_name'][1:]])
            d['c1i'] = ''.join(['{:3d}'.format(ci) for ci in d['chopper_index'][1:]])
            d['coil'] = coil
            output += [out_format.format(**d)]
        if not self.filename.startswith('patnow'):
            output += ['Configuration Description']
        output = '\n'.join(output)
        self.contents = output
        return output

    def _write_type_i(self):
        """Method for forming output string for type-I patch panel files"""
        output = []
        output += ['C-coil parity: ' + self.get('config', {}).get('c_coil_parity', '')]
        blank_c = dict(slot='', sign='', spa='', coil='', c_idx=0)
        for i in range(10):
            output += [
                str(1 + int(np.floor(i / 2)))
                + ' {slot:2s} {sign:3s} {spa:>4s} {c_idx:1d} {coil:58s}'.format(**self.get('c{}'.format(i), blank_c))
            ]
        output += ['I-Coil file: ' + self.get('config', {}).get('i_coil_file', '')]
        for a in range(1, 13):
            output += ['AA{:<2d} {:69s}'.format(a, self.get('AA', {}).get('AA{}'.format(a), ''))]
        for i in [1, 2]:
            for a in 'ABCDEF':
                ss = 'SS{}{}'.format(i, a)
                output += ['{:6s} {:67s}'.format(ss, self.get('SS', {}).get(ss, ''))]
        output = '\n'.join(output)
        self.contents = output
        return output

    @dynaSave
    def save(self, no_write=False):
        """Saves file to disk"""
        if self.patch_type == 'F':
            output = self._write_type_f()
        elif self.patch_type == 'P':
            raise TypeError(
                'Cannot save type P files; set self.patch_type="F" first to allow saving. Type P is F-coil patch in '
                'some old format, which is binary and hard to use. Type F is F-coil patch in archival format, which is '
                'ASCII and easy to deal with. If a type-P file is converted, it is recommended that a C-coil entry be '
                'added using self.add_c_coil(). Specify d_supply=True if the D power supply should power C-coils.'
            )
        elif self.patch_type == 'I':
            output = self._write_type_i()
        else:
            raise NotImplementedError('Patch type {} does not have a write method yet; sorry'.format(self.patch_type))
        if not no_write:
            with open(self.filename, 'w') as f:
                self.printq('Writing patch panel main file {}...'.format(self.filename))
                f.write(output)
            if os.path.split(self.filename)[1].endswith('patnow.dat') and (self.patch_type in 'FI'):
                name_file = self.filename.replace('patnow.dat', 'patnow.nam')
                name_output = '\n'.join([self.patch_name, 'Configuration Description'])
                self.printq('Writing patch panel name file {}...'.format(name_file))
                with open(name_file, 'w') as f:
                    f.write(name_output)
        return output



import sys
from tkinter import *
import tkinter.font as tkFont
from tkinter import ttk
import re
import copy

root = Tk()

top = Frame(root)
top.grid(row=0, column=0)
canvas = Canvas(top)
vbar = Scrollbar(top, orient=VERTICAL)
vbar.pack(side=RIGHT, fill=Y)
vbar.config(command=canvas.yview)
canvas.config(yscrollcommand=vbar.set)
canvas.pack(side=LEFT, expand=True, fill=BOTH)

frame = Frame()
sys_fonts = list(tkFont.families(root))
sys_fonts.sort()

_tmp = Label(text="bla")
f = tkFont.Font(font=_tmp['font'])
_tmp.destroy()
font0 = f.actual()
for k in list(font0.keys()):
    if k not in ['family', 'weight', 'slant']:
        del font0[k]
font0['family'] = re.sub(' ', '\ ', font0['family'])

fonts = []
font = copy.deepcopy(font0)

font1 = copy.deepcopy(font)
fonts.append(tkFont.Font(**font1))

font2 = copy.deepcopy(font)
font2['slant'] = 'italic'
fonts.append(tkFont.Font(**font2))

font3 = copy.deepcopy(font)
font3['weight'] = 'bold'
fonts.append(tkFont.Font(**font3))

Label(frame, text='default: ' + f['family'], font=fonts[0]).pack(side=TOP, fill=X)
Label(frame, text='default: ' + f['family'], font=fonts[1]).pack(side=TOP, fill=X)
Label(frame, text='default: ' + f['family'], font=fonts[2]).pack(side=TOP, fill=X)
ttk.Separator(frame).pack(side=TOP, fill=X)

for item in sys_fonts:
    try:
        font = copy.deepcopy(font0)
        font['family'] = re.sub(' ', '\ ', item)

        font1 = copy.deepcopy(font)
        fonts.append(tkFont.Font(**font1))

        font2 = copy.deepcopy(font)
        font2['slant'] = 'italic'
        fonts.append(tkFont.Font(**font2))

        font3 = copy.deepcopy(font)
        font3['weight'] = 'bold'
        fonts.append(tkFont.Font(**font3))

        Label(frame, text=item, font=fonts[-3]).pack(side=TOP, fill=X)
        Label(frame, text=item, font=fonts[-2]).pack(side=TOP, fill=X)
        Label(frame, text=item, font=fonts[-1]).pack(side=TOP, fill=X)

        ttk.Separator(frame).pack(side=TOP, fill=X)
        print('OK', item)
    except Exception as _excp:
        print('NO', item, _excp)

canvas.create_window(0, 0, anchor=NW, window=frame)
canvas.update_idletasks()
top.configure(width=top.winfo_reqwidth())
canvas.configure(width=frame.winfo_reqwidth(), height=500)
canvas.configure(scrollregion=(0, 0, top.winfo_width(), frame.winfo_reqheight()))

root.mainloop()



# -*-Python-*-
# Created by eldond at 2019-12-23 08:43

"""
This script produces a GUI for managing EFIT-viewer style plots
"""

from omfit_classes.omfit_efitviewer import (
    omas_features,
    plot_cx_ods,
    load_cx_ods,
    update_cx_time,
    plot_grid,
    dump_plot_command,
    omas_features,
    grab_xylim,
    default_xylim,
    update_overlay,
    add_case,
    delete_case,
    setup_avail_systems,
    setup_special_contours,
    load_contours,
    pick_output_location,
    box_plot_command,
    get_default_snap_list,
    EfitviewerDataError,
    populate_efitviewer_figure,
    save_subplot_layout,
    find_cq_options,
    find_profile_y_options,
    setup_default_profiles,
    update_gui_figure,
)

from omfit_classes.omfit_efitviewer import update_gui_figure as update_figure
from omfit_classes.utils_fusion import available_EFITs

defaultVars(section='top_level', index=0, location=None, hw_sys=None)

# Link to related items
that = MainScratch['__efitviewer_support_gui__']
out_loc, settings_loc, out, settings, parent_settings, scratch = pick_output_location()
settings.setdefault('profiles', SettingsName())

if index == 0:
    ods_tag = 'efitviewer_ods'
else:
    ods_tag = 'efitviewer_ods_{}'.format(index)

if ods_tag in out:
    re_load = 'Reload (will overwrite current case) '
else:
    re_load = 'Load'


def top_level():
    """Top level of the efitviewer mk.2 GUI"""
    OMFITx.TitleGUI('efitviewer mk.2')

    with OMFITx.same_row():
        OMFITx.CompoundGUI(this, '', section='efitviewer_controls')
        OMFITx.CompoundGUI(this, '', section='efitviewer_figure')
    return


def efitviewer_figure(show_labels=True):
    """
    Provides the figure panel to the eftiviewer GUI

    :param show_labels: bool
        Enable an OMFITx.Label instance.

    :return: (Axes instance, OMFITx.Label() result)
    """
    label = None

    show_cx = settings.setdefault('show_cx', True)
    num_profiles = settings['profiles'].setdefault('num_profiles', 0)
    num_profile_cols = settings['profiles'].setdefault('num_profile_cols', 0)
    gridspec_kw = settings['profiles'].setdefault('gridspec_kw', {})

    if 'efitviewer_ods' in out:
        ods = out['efitviewer_ods']
        if 'dataset_description.data_entry.machine' in ods and 'dataset_description.data_entry.pulse' in ods:
            device = out['efitviewer_ods']['dataset_description.data_entry.machine']
            shot = out['efitviewer_ods']['dataset_description.data_entry.pulse']
            if show_labels:
                label = OMFITx.Label('{}#{}'.format(device, shot))
        elif show_labels:
            label = OMFITx.Label('Incomplete ODS. Please try reloading or loading a different case.')
    elif show_labels:
        label = OMFITx.Label('Waiting for ODS data to be loaded...')

    fig = scratch['efitviewer_fig'] = OMFITx.Figure(returnFigure=True, figsize=settings.setdefault('figsize', [5, 8]))
    cx_ax, profile_axs = populate_efitviewer_figure(
        fig,
        scratch,
        show_cx=show_cx,
        num_profiles=num_profiles,
        num_profile_cols=num_profile_cols,
        profiles_sharex=settings['profiles'].setdefault('sharex', 'all'),
        profiles_sharey=settings['profiles'].setdefault('sharey', 'none'),
        **gridspec_kw,
    )
    OMFITx.Entry(settings_loc + "['figsize']", 'Figure dimensions (cm)', default=[7, 8], updateGUI=True, width=5)
    with OMFITx.same_row():
        OMFITx.Button(
            'Save subplot layout',
            save_subplot_layout,
            width=5,
            help='Getting the right subplot spacing when the figure has a cross section and an arbitrary number '
            'of profile subplots is hard to do automatically. You can use the adjustment sliders to make '
            'the figure how you want it and then save your layout with this button. Each combination of cross '
            'section on/off, number of profiles, and number of columns for profiles will get a unique save, so '
            'you can switch back and forth between profile plot counts and your settings should be remembered.',
        )
        OMFITx.Button(
            'Forget saved layout',
            lambda: save_subplot_layout(reset=True),
            updateGUI=True,
            width=5,
            help='Delete a saved subplot layout so it will return to the global default settings. '
            'Layouts are saved for each combination of CX on/off, number of profile plots, and number of '
            'columns of profile plots, so you can save a different left margin for a figure with just a '
            'cross section from the left margin for a figure with 9 profiles in 3 columns in addition to '
            'the cross section..',
        )
    # OMFITx.CheckBox(settings_loc + "['auto_plot']", 'Automatic update', default=True, updateGUI=True)
    if 'efitviewer_ods' in out:
        if settings['auto_plot']:
            plot_cx_ods(clear_first=True, gentle=True)  # plot_cx_ods will get fig & ax from settings in scratch
        else:
            OMFITx.Button('Plot cross section', lambda: plot_cx_ods(clear_first=True, gentle=True))
    return cx_ax, label


def efitviewer_controls():
    """Provides the main control part of the efitviewer GUI"""

    # Choose G-file or device/shot/tree/time
    OMFITx.CompoundGUI(this, '', section='efitviewer_multi_case_selector')
    # The time overlay checkbox should be outside of the case-specific tabs
    OMFITx.CheckBox(settings_loc + "['overlay']", 'Overlay times', default=False, postcommand=update_overlay)

    if 'efitviewer_ods' in out:
        ods = out['efitviewer_ods']
        ods_valid = 'time' in ods['equilibrium'] and len(ods['equilibrium']['time_slice'][0]['profiles_2d'])
        if ods_valid:
            OMFITx.CompoundGUI(this, '', section='efitviewer_overlays')  # Choose hardware system overlays
            OMFITx.CompoundGUI(this, '', section='efitviewer_details')  # Stuff like figure setup, grid, etc.
        else:
            OMFITx.Label('ODS appears to be invalid. Try troubleshooting, then reloading.', fg='red')
            OMFITx.CompoundGUI(this, 'Troubleshooting tools', section='efitviewer_troubleshooting')
    else:
        OMFITx.Tab('Advanced')
        OMFITx.CompoundGUI(this, '', section='efitviewer_advanced')
    return


def efitviewer_advanced():
    """Advanced options"""
    OMFITx.CheckBox(
        settings_loc + "['no_empty']",
        'Remove empty time-slices from ODS',
        default=True,
        help='Inspect plasma current and number of boundary outline points at each time and '
        'remove time-slices where either of these is 0. Takes effect before MDSplus EFIT data are written to ODS.',
    )
    OMFITx.Entry(
        settings_loc + "['int_time_roundoff_threshold']",
        'Threshold for rounding times to integers (ms)',
        default=1e-2,
        help='Do you want to see 326 ms instead of 326.000000123 ms? Well, this is the setting for you!',
        updateGUI=True,
    )


def efitviewer_multi_case_selector():
    """Provides a sub-GUI for managing several case selectors for doing multi-shot overlays"""

    OMFITx.Tab('Primary case')
    OMFITx.CompoundGUI(this, '', section='efitviewer_case_selector', index=0)

    cases = settings['cases']

    for i in range(1, len(cases)):
        OMFITx.Tab('Case {}'.format(i))
        OMFITx.CompoundGUI(this, '', section='efitviewer_case_selector', index=i)
        OMFITx.Button('Delete this case', lambda idx=i: delete_case(idx), updateGUI=True)

    OMFITx.Tab('New case')
    OMFITx.Button('Add case', add_case, updateGUI=True)
    return


def time_selector(ods_tag):
    """
    Adds the time selector
    """
    try:
        efit_times = out[ods_tag]['equilibrium.time'] * 1000  # s to ms
    except (KeyError, ValueError):
        efit_times = None
    if efit_times is not None and len(tolist(efit_times)):
        rounded_times = np.around(efit_times).astype(int)
        integer_conversion_error = max(abs(efit_times - rounded_times))  # ms
        int_time_roundoff_threshold = settings.setdefault('int_time_roundoff_threshold', 1e-2)
        if integer_conversion_error < int_time_roundoff_threshold:  # ms
            efit_times = rounded_times
        if parent_settings['EXPERIMENT']['time'] is None:
            default_time = efit_times[0]
        else:
            default_time = efit_times[closestIndex(efit_times, parent_settings['EXPERIMENT']['time'])]
    else:
        default_time = None
    if index == 0:
        case_name = 'primary case'
    else:
        case_name = f'case {index}'

    def slider2combobox_sync(location):
        update_cx_time(location)
        time_combobox.set(eval(location))

    def combobox2slider_sync(location):
        update_cx_time(location)
        time_slider.scale.set(eval(location))

    time_slider = OMFITx.Slider(
        "scratch['efitviewer_{}_time']".format(index),
        [efit_times[0], efit_times[-1], 1],
        "Time (ms)",
        digits=0,
        default=default_time,
        postcommand=slider2combobox_sync,
        refresh_every=1,
    )
    time_combobox = OMFITx.ComboBox(
        "scratch['efitviewer_{}_time']".format(index),
        tolist(efit_times),
        "Time (ms)",
        default=default_time,
        check=is_int,
        state='normal',
        postcommand=combobox2slider_sync,
    )
    # t = settings['cases'][index].get('time', None)
    if settings['cases'][index].setdefault('time', None) is None:
        # noinspection PyBroadException
        try:
            settings['cases'][index]['time'] = scratch[f'efitviewer_{index}_time']
        except Exception:
            pass
    # OMFITx.CheckBox(settings_loc + "['overlay']", 'Overlay times', default=False, postcommand=update_overlay)


def add_status_label(index, data_source, show_status_label=True, color=None, text=None):
    """
    Adds a status label to the GUI

    :param index: int
        Index of this case, for differentiating between tabs for multi-shot display.

    :param data_source: str
        'MDS': data from MDSplus
        'tree': data from an object in the tree
        other: raises EfitviewerDataError

    :param show_status_label: bool
        Flag for enabling the label

    :param color: str
        Name of a color for the label foreground. If provided, it will be applied.

    :param text: str
        New text to use to update the label. If not provided, the label text will remain unchanged.
    """
    # In order to work properly, the content of 'select_text_var_#' in scratch has to be set by the client
    select_text_var = scratch.setdefault(f'select_text_var_{index}_{data_source}', tk.StringVar())

    if show_status_label:
        scratch[f'efitviewer_select_label_{index}_{data_source}'] = OMFITx.Label('', textvariable=select_text_var)
        update_status_label(index, data_source, color=color, text=text)
    else:
        scratch[f'efitviewer_select_label_{index}_{data_source}'] = None


def update_status_label(index, data_source, color=None, text=None):
    """
    Updates the status label

    :param index: int
        Index of this case, for differentiating between tabs for multi-shot display.

    :param data_source: str
        'MDS': data from MDSplus
        'tree': data from an object in the tree
        other: raises EfitviewerDataError

    :param color: str
        Name of a color for the label foreground. If provided, it will be applied.

    :param text: str [optional]
        New text to use to update the label. If not provided, the label text will remain unchanged.
    """
    label = scratch.setdefault(f'efitviewer_select_label_{index}_{data_source}', None)
    select_text_var = scratch.setdefault(f'select_text_var_{index}_{data_source}', tk.StringVar())
    if select_text_var is not None and text is not None:
        select_text_var.set(text)
    if label is not None:
        if color is not None:
            label.config(foreground=color)


def efitviewer_mds_selector(show_status_label=True):
    """
    Sets up the source of data from MDSplus

    :param show_status_label: bool
    """
    data_source = 'MDS'
    OMFITx.Entry(
        settings_loc + "['cases'][{}]['device']".format(index),
        'Device',
        check=is_string,
        default=tokamak(parent_settings['EXPERIMENT']['device']),
        updateGUI=True,
    )
    OMFITx.Entry(
        settings_loc + "['cases'][{}]['shot']".format(index),
        'Shot',
        check=is_int,
        default=evalExpr(parent_settings['EXPERIMENT']['shot']),
        updateGUI=True,
    )
    device = tokamak(settings['cases'][index]['device'])
    shot = settings['cases'][index]['shot']

    snap_list, efit_help = available_EFITs(scratch, device, shot, default_snap_list=get_default_snap_list(device), format='{tree}')

    e = OMFITx.ComboBox(
        settings_loc + "['cases'][{}]['efit_tree']".format(index),
        snap_list,
        'EFIT tree',
        state='normal',
        help=efit_help,
        updateGUI=True,
        default=list(snap_list.values())[0] if (len(snap_list) > 0) else None,
    )
    e.configure(width=40)
    efit_tree = settings['cases'][index]['efit_tree']

    OMFITx.CheckBox(
        f"{settings_loc}['minimal_eq_data']",
        "Load only minimal data",
        default=True,
        help='A minimal dataset should load faster, but will only support the cross section plot. '
        'Options for profile plots will be very limited.',
    )

    scratch[f'select_text_var_{index}_{data_source}'] = select_text_var = tk.StringVar()
    try:
        ods_device = out[ods_tag]['dataset_description.data_entry.machine']
        ods_shot = out[ods_tag]['dataset_description.data_entry.pulse']
        ods_comment = out[ods_tag]['dataset_description.ids_properties.comment']
    except (KeyError, ValueError):
        ods_device = None
        ods_shot = None
        ods_efit_tree = None
        ods_ready = False
    else:
        ods_efit_tree = ods_comment.split('EFIT tree = ')[-1]
        ods_ready = True
    if (efit_tree is None) or (shot is None) or (device is None):
        select_text_var.set('Select valid device, shot, and tree to proceed')
        select_text_color = 'black'
    elif not ods_ready:
        select_text_var.set('Ready to load')
        select_text_color = 'black'
    elif (shot != ods_shot) or (device != ods_device) or (efit_tree != ods_efit_tree):
        select_text_var.set('Press reload to update ODS to match new settings.')
        select_text_color = 'red'
    else:
        select_text_var.set('Current settings match ODS; ready to plot!')
        select_text_color = 'green'

    add_status_label(index, data_source, show_status_label, color=select_text_color)
    if efit_tree is not None and shot is not None and device is not None:
        OMFITx.Button(
            '{} {}#{} {}'.format(re_load, device, shot, efit_tree),
            lambda: load_cx_ods(gfile='unused', index=index),
            updateGUI=True,
        )
    return ods_tag


def compare_gfile_to_ods(gfile, ods, require_shot=True, require_efitid=False, require_device=False, require_times=True):
    """
    Compares as much information as possible in a gfile to that in an ods.

    Used for deciding how to update the status label when selecting a data source from the tree.

    :param gfile: dict-like, or OMFITgeqdsk
        A single OMFITgeqdsk can be used directly.
        A dict-like object, like an OMFITcollection or OMFITtree, must contain at least one OMFITgeqdsk instance.

    :param ods: ODS instance

    :param require_shot: bool
        Fail if shot can't be found for either source. Shot can usually be found in both.

    :param require_efitid: bool
        Fail if EFIT ID can't be found for either source. This isn't always easy to look up procedurally.

    :param require_device: bool
        Fail if device can't be found for either source. This isn't always easy to look up procedurally.

    :param require_times: bool
        Fail if times can't be found for either source. They are easy to find.

    :return: bool
        Flag indicating whether or not the gfile and ODS are consistent with each other.
    """

    ods_device = ods['dataset_description.data_entry.machine']
    ods_shot = ods['dataset_description.data_entry.pulse']
    ods_efitid = None
    ods_times = ods['equilibrium.time']

    # Get gfile times and a reference to a single OMFITgeqdsk instance
    if isinstance(gfile, OMFITgeqdsk):
        gfile0 = gfile
        gfile_times = np.array([int(gfile0.filename.split('.')[-1])]).astype(float)
    elif isinstance(gfile, dict):
        gfile_keys = [k for k, v in gfile.items() if isinstance(v, OMFITgeqdsk)]
        if not len(gfile_keys):
            raise EfitviewerDataError("This dictionary does not include any OMFITgeqdsk instances")
        gfile0 = gfile[gfile_keys[0]]
        gfile_times = np.array(gfile_keys).astype(float)
    else:
        raise EfitviewerDataError("gfile input to compare_gfile_to_ods() must be an OMFITgeqdsk or a dict-like")

    if (ods_times is not None) and (np.max(gfile_times) >= (900 * np.max(ods_times))):
        gfile_times *= 1e-3  # ms to s

    # Try to figure out geqdsk shot number
    if 'g' in gfile0.filename and '.' in gfile0.filename:
        try:
            gfile_shot_fn = int(gfile0.filename.split('g')[1].split('.')[0])
        except (ValueError, IndexError):
            gfile_shot_fn = None
    else:
        gfile_shot_fn = None
    try:
        gfile_shot_case = int(gfile0['CASE'][3].split('#')[1])
    except (IndexError, ValueError, KeyError):
        gfile_shot_case = None
    if gfile_shot_case is not None:
        gfile_shot = gfile_shot_case
    else:
        gfile_shot = gfile_shot_fn

    # Try to figure out geqdsk efitid
    try:
        gfile_efitid = gfile0['CASE'][5].strip()
    except (KeyError, IndexError, AttributeError):
        gfile_efitid = None

    # I don't know a standard for recording this
    gfile_device = None

    # Do the comparisons
    def do_comparison(gg, oo, rr, tag=''):
        if gg is not None and oo is not None:
            mm = not array_equal(gg, oo)
            printd(f'{ ["no mismatch", "mismatch"][int(mm)]} between {gg} and {oo}   {tag}', topic='efitviewer')
        elif rr:
            printd(f'mismatch between {gg} and {oo} because one is missing   {tag}', topic='efitviewer')
            mm = True
        else:
            printd(f'no mismatch between {gg} and {oo} because missing data are ignored   {tag}', topic='efitviewer')
            mm = False
        return mm

    device_mismatch = do_comparison(gfile_device, ods_device, require_device)
    shot_mismatch = do_comparison(gfile_shot, ods_shot, require_shot)
    efitid_mismatch = do_comparison(gfile_efitid, ods_efitid, require_efitid)
    times_mismatch = do_comparison(gfile_times, ods_times, require_times)
    return not (device_mismatch + shot_mismatch + efitid_mismatch + times_mismatch)


def check_eqdsk_type(xeqdsk, desired_type=None, allow_contains=True):
    """
    Checks that an object is an *EQDSK file (g/a/m/k) or a container that contains EQDSKs

    :param xeqdsk: object
        Object to test

    :param desired_type: type
        OMFITgeqdsk, OMFITaeqdsk, OMFITmeqdsk, or OMFITkeqdsk.
        Supply None to accept any of these types.

    :param allow_contains: bool
        True: The object is valid if it's container that contains the right kind of file.
        False: The object iself must be the right kind of file.

    :return: bool
        xeqdsk is valid according to the desired type and container rules
    """

    if desired_type is None:
        desired_type = [OMFITgeqdsk, OMFITmeqdsk, OMFITaeqdsk, OMFITkeqdsk]

    x_valid = isinstance(xeqdsk, desired_type)
    if allow_contains and isinstance(xeqdsk, dict):
        xfiles = [k for k, v in xeqdsk.items() if isinstance(v, desired_type)]
        x_valid = x_valid or bool(len(xfiles))
    return x_valid


def efitviewer_tree_selector(show_status_label=True):
    """
    Setup for a data source in the OMFIT tree

    :param show_status_label: bool
    """
    from omfit_classes.omfit_omas_utils import check_shot_time

    data_source = 'tree'
    for gamk in 'gamk':
        gamk_io = 'INPUTS' if gamk == 'k' else 'OUTPUTS'
        try:
            gamk_time0 = root[gamk_io]['gEQDSK'].keys()[0]
            default_gamk_file = f"root['{gamk_io}']['{gamk}EQDSK']['{gamk_time0}']"
        except (KeyError, IndexError):
            default_gamk_file = 'unused'
        OMFITx.TreeLocationPicker(
            f"{settings_loc}['cases'][{index}]['{gamk}file']",
            f'{gamk.upper()}-file(s) in tree',
            default=default_gamk_file,
            updateGUI=True,
            help='G-files provide the essential data for plotting the equilibrium. A-files contain additional '
            'quantities. M-files contain measurements that can be used as inputs as well as reconstructed values '
            'to compare to those measaurements and chi-squared values. K-files contain constraints and setup '
            'instructions.',
        )

    is_g = contains_g = None
    try:
        gfile_str = settings['cases'][index]['gfile']
        gfile = eval(settings['cases'][index]['gfile'])
        is_g = isinstance(gfile, OMFITgeqdsk)
        if isinstance(gfile, dict):
            gfiles = [k for k, v in gfile.items() if isinstance(v, OMFITgeqdsk)]
            contains_g = bool(len(gfiles))
        else:
            contains_g = False
            gfiles = None
        gfile_valid = is_g or contains_g
    except (KeyError, IndexError, TypeError, NameError):
        gfile_str = None
        gfile_valid = False
        case_info = ''
    else:
        if is_g:
            case_info = ' '.join(gfile['CASE'])
            shot = int(gfile['CASE'][3].split('#')[-1])
            t = float(gfile['CASE'][4].split('ms')[0])
        elif contains_g:
            case_info = ' '.join(array(gfile[gfiles[0]]['CASE'])[array([0, 1, 2, 3, 5])])
            shot = int(gfile[gfiles[0]]['CASE'][3].split('#')[-1])
            t = np.array([float(gf['CASE'][4].split('ms')[0]) for gf in gfile.values()])
        else:
            case_info = ''
            shot = None
            t = None
    if gfile_valid:
        try:
            afile = eval(settings['cases'][index]['afile'])
        except (KeyError, IndexError, TypeError, NameError):
            afile = None
            afile_valid = False
        else:
            afile_valid = check_shot_time(shot, t, afile, raise_on_bad=False) and check_eqdsk_type(afile, OMFITaeqdsk)

        try:
            mfile = eval(settings['cases'][index]['mfile'])
        except (KeyError, IndexError, TypeError, NameError):
            mfile = None
            mfile_valid = False
        else:
            mfile_valid = check_shot_time(shot, t, mfile, raise_on_bad=False) and check_eqdsk_type(mfile, OMFITmeqdsk)

        try:
            kfile = eval(settings['cases'][index]['kfile'])
        except (KeyError, IndexError, TypeError, NameError):
            kfile = None
            kfile_valid = False
        else:
            kfile_valid = check_shot_time(shot, t, kfile, raise_on_bad=False) and check_eqdsk_type(kfile, OMFITkeqdsk)

        not_included = []

        if afile_valid:
            case_info += '\n+aEQDSK data'
        else:
            not_included += ['aEQDSK']
        if mfile_valid:
            case_info += '\n+mEQDSK data'
        else:
            not_included += ['mEQDSK']
        if kfile_valid:
            case_info += '\n+kEQDSK data'
        else:
            not_included += ['kEQDSK']
        if len(not_included):
            not_included = 'Not included: ' + ', '.join(not_included)
    else:
        afile_valid = mfile_valid = kfile_valid = None
        not_included = []

    ods_loaded = ods_tag in out

    scratch[f'select_text_var_{index}_{data_source}'] = select_text_var = tk.StringVar()
    try:
        ods_source = out[ods_tag]['dataset_description.ids_properties.source']
    except (KeyError, ValueError):
        ods_source = None
    if (ods_source == gfile_str) and gfile_valid:
        if compare_gfile_to_ods(gfile, out[ods_tag]):
            select_text_var.set('Current settings match ODS; ready to plot!')
            select_text_color = 'green'
        else:
            select_text_var.set('Current GEQDSK location matches ODS, but a discrepancy was detected.')
            select_text_color = 'red'
    elif ods_loaded and gfile_valid:
        select_text_var.set('Press reload to update ODS to match new settings.')
        select_text_color = 'red'
    elif not ods_loaded and gfile_valid:
        select_text_var.set('Ready to load.')
        select_text_color = 'black'
    else:
        select_text_var.set('Select an OMFITgeqdsk or container of OMFITgeqdsk instances to continue.')
        select_text_color = 'red'

    add_status_label(index, data_source, show_status_label, color=select_text_color)
    if gfile_valid:
        OMFITx.Button('{} {}'.format(re_load, case_info), lambda: load_cx_ods(index=index), updateGUI=True)
        if len(not_included):
            OMFITx.Label(not_included, fg='red')
    # else:
    # OMFITx.Label('Select an OMFITgeqdsk or container of OMFITgeqdsk instances to continue')
    return ods_tag


def efitviewer_case_selector(show_status_label=True):
    """
    Provides a sub-GUI with controls for choosing (device, shot, tree, time) or a g-file for efitviewer-style plots

    :param show_status_label: bool
        Add an OMFITx.Label() with the current status.
    """

    if index not in settings['cases']:
        OMFITx.Label(f"Index {index} is not ready. Possible corruption in settings['cases']", fg='red')
        OMFITx.End()

    OMFITx.ComboBox(
        f"{settings_loc}['cases'][{index}]['source']",
        {'Gather data from MDSplus': 'MDS', 'Gather a GEQDSK or collection of GEQDSKs from the tree': 'tree'},
        'Data source',
        default='MDS',
        updateGUI=True,
    )
    src = settings['cases'][index]['source']
    fn_name = f'efitviewer_{src.lower()}_selector'
    fn = eval(fn_name)
    fn(show_status_label=show_status_label)

    # Done with gEQDSK selection ----------------------------------------------------------------------------------
    OMFITx.Tab()

    if ods_tag in out:
        with OMFITx.same_row():
            OMFITx.CheckBox(
                settings_loc + "['cases'][{}]['active']".format(index),
                'Overlay equilibrium from this case',
                default=True,
                postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
            )
            OMFITx.Button(
                'Customize equilibrium...',
                lambda location=None, idx=index: that.run(section='efitviewer_equilibrium_customization', index=idx, out_loc=out_loc),
            )

    if ods_tag in out:
        time_selector(ods_tag)

    return


def efitviewer_overlays():
    """Provides a GUI panel for selecting hardware overlays for the efitviewer-style plot"""

    try:
        device = out['efitviewer_ods']['dataset_description.data_entry.machine']
    except (ValueError, KeyError):
        device = tokamak(parent_settings['EXPERIMENT']['device'])
    setup_avail_systems()
    for i, sys in enumerate(settings['systems'].keys()):
        if (i % 10) == 0:
            OMFITx.Tab('Overlays {}'.format(i))
        with OMFITx.same_row():
            OMFITx.CheckBox(
                settings_loc + "['systems']['{}']".format(sys),
                sys,
                default=False,
                postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
            )
            OMFITx.Button(
                'Customize...', lambda sys_=sys: that.run(hw_sys=sys_, section='co_efv_{}'.format(sys_), out_loc=out_loc), width=5
            )

    OMFITx.Tab('Overlays special')
    OMFITx.CompoundGUI(this, '', section='efv_so_general')

    dev_tag = 'efv_so_{}'.format(tokamak(device, output_style='GPEC'))
    OMFITx.Tab('Overlays {}'.format(device))
    try:
        OMFITx.CompoundGUI(this, '', section=dev_tag)
    except NameError:
        OMFITx.Label('No special instructions are defined for this device.')

    OMFITx.Tab('Profiles')  # -------------------------------------------------------------------------------------
    OMFITx.CompoundGUI(this, '', section='efitviewer_profile_setup')
    return


def efitviewer_details():
    """Provides a sub-GUI for changing details of the efitviewer tool"""

    # Setup  # ----------------------------------------------------------------------------------------------------
    try:
        device = out['efitviewer_ods']['dataset_description.data_entry.machine']
        shot = out['efitviewer_ods']['dataset_description.data_entry.pulse']
    except (KeyError, ValueError):
        device = tokamak(parent_settings['EXPERIMENT']['device'])
        shot = evalExpr(parent_settings['EXPERIMENT']['shot'])

    OMFITx.Tab('Axes')  # -----------------------------------------------------------------------------------------
    if True:  # For code folding, okay?
        with OMFITx.same_row():
            lim_help = 'Select axis limits for the plot in data units (probably m).'
            grab_help = 'Grab current axis limits from the figure as displayed and put them into the entry fields.'
            default_help = 'Set axis limits to reasonable defaults for the current device, if there are any on file.'
            # If using same_row: attach help to last item on the row. Otherwise, split it up.
            last_help = 'Control plot axis limits.\n\nEntry fields: {}\n\nGrab button: {}\n\nd button: {}'.format(
                lim_help, grab_help, default_help
            )
            OMFITx.Label('Limits (m):')
            OMFITx.Entry(settings_loc + "['xlim']", 'R', width=9, postcommand=update_figure)
            OMFITx.Entry(settings_loc + "['ylim']", 'Z', width=9, postcommand=update_figure)
            OMFITx.Button('Grab', grab_xylim, updateGUI=True, width=5)
            OMFITx.Button('d', default_xylim, updateGUI=True, width=1, help=last_help)

        with OMFITx.same_row():
            settings.setdefault('plot_style_kw', {}).setdefault('axes_aspect', 'equal_box')
            settings.setdefault('plot_style_kw', {}).setdefault('frame_on', True)

            OMFITx.ComboBox(
                settings_loc + "['plot_style_kw']['axes_aspect']",
                ['equal_box', 'equal_datalim', 'auto', 'None'],
                'Aspect',
                # default='equal_box',
                state='normal',
                postcommand=update_figure,
                width=6,
                help='"equal_box": forces even X/Y axis spacing by changing the axes dimensions. '
                '"equal_datalim": forces even X/Y axis spacing by changing data limits. '
                '"auto": allows uneven X/Y axis spacing. '
                'numeric: a circle will be stretched so its height is X times its width, where X is this setting. '
                'None: no action; aspect will be determined by some other script, such as a plot call in OMAS.',
            )
            OMFITx.CheckBox(settings_loc + "['plot_style_kw']['frame_on']", 'Frame')

        with OMFITx.same_row():
            settings.setdefault('plot_style_kw', {}).setdefault('tick_spacing', 0.25)
            settings.setdefault('plot_style_kw', {}).setdefault('xtick_loc', 'both')
            settings.setdefault('plot_style_kw', {}).setdefault('ytick_loc', 'both')
            OMFITx.Label('Ticks:')
            OMFITx.Entry(
                settings_loc + "['plot_style_kw']['tick_spacing']",
                'spacing (m)',
                # default=0.25,  # Default buttons make the GUI too wide
                postcommand=update_figure,
                width=5,
            )
            OMFITx.Label('Position:')
            OMFITx.ComboBox(
                settings_loc + "['plot_style_kw']['xtick_loc']",
                ['top', 'bottom', 'both'],
                'X',
                # default='both',
                postcommand=update_figure,
                width=6,
            )
            OMFITx.ComboBox(
                settings_loc + "['plot_style_kw']['ytick_loc']",
                ['left', 'right', 'both'],
                'Y',
                # default='both',
                postcommand=update_figure,
                width=6,
            )

        with OMFITx.same_row():
            OMFITx.CheckBox(
                settings_loc + "['plot_style_kw']['grid_enable']",
                'Show grid',
                default=False,
                postcommand=lambda location: plot_grid(enable=eval(location)),
            )
            OMFITx.Entry(
                settings_loc + "['plot_style_kw']['grid_kw']",
                'Grid keywords',
                default={},
                postcommand=lambda location: plot_grid(grid_kw=eval(location)),
                help='Dictionary of keywords to pass to grid(). '
                'Some options: color, linestyle, linewidth, alpha, ... . '
                'For more, see https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.grid.html',
            )

    OMFITx.Tab('Annotations')  # ----------------------------------------------------------------------------------
    if True:
        OMFITx.Entry(
            settings_loc + "['default_fontsize']",
            'OVERRIDE default annotation fontsize',
            default=None,
            help='12 is the default OMFIT size for the default OMFIT style. The current setting of '
            'rcParams["font.size"] is {}. Set this to None to leave rcParams alone. '
            'OMAS overlays specify relative font sizes by default, like xx-small or medium. '
            'These will scale with rcParams["font.size"].'.format(rcParams['font.size']),
            postcommand=update_figure,
        )
        OMFITx.ComboBox(
            settings_loc + "['show_legend']", {'always': 1, 'automatic': 2, 'never': 0}, 'Show legend', default=2, postcommand=update_figure
        )
        OMFITx.ComboBox(
            settings_loc + "['show_cornernote']",
            {'always': 1, 'automatic': 2, 'never': 0},
            'Show cornernote',
            default=2,
            postcommand=update_figure,
        )
        with OMFITx.same_row():
            settings.setdefault('plot_style_kw', {}).setdefault('subplot_label', None)
            settings.setdefault('plot_style_kw', {}).setdefault('subplot_label_corner', [1, 1])

            OMFITx.ComboBox(
                settings_loc + "['plot_style_kw']['subplot_label']",
                {'': None, '(a)': 0, '(b)': 1, '(c)': 2, '(d)': 3, '(e)': 4, '(f)': 5, '(g)': 6, '(h)': 7},
                'Add subplot label',
                width=5,
                postcommand=update_figure,
            )
            OMFITx.ComboBox(
                settings_loc + "['plot_style_kw']['subplot_label_corner']",
                {'top right': [1, 1], 'top left': [0, 1], 'bottom right': [1, 0], 'bottom left': [0, 0]},
                '',
                postcommand=update_figure,
                width=5,
            )

    OMFITx.Tab('Contours')  # -------------------------------------------------------------------------------------
    if True:
        if compare_version(omas.__version__, omas_features['contour_quantity']) >= 0:
            cq_opts, default_contour_quantity = find_cq_options()
            if len(cq_opts):
                OMFITx.ComboBox(
                    settings_loc + "['contour_quantity']",
                    cq_opts,
                    'Contour quantity',
                    default=default_contour_quantity,
                    updateGUI=True,
                    state='normal',
                )
            else:
                OMFITx.Label('No phi or psi data found in equilibrium 2d profiles! Contours unavailable!', fg='red')
                settings.setdefault('contour_quantity', 'rho')  # Make sure it's defined
        else:
            # Wasn't controlled well in OMAS 0.49.1 and earlier, but the default if all data are available was rho
            settings['contour_quantity'] = 'rho'

        bcl = np.r_[0.1:10:0.1]
        default_contour_levels = {'q': [1, 2, 2.5, 3, 4, 5, 6]}.get(settings['contour_quantity'], bcl)

        OMFITx.Entry(
            settings_loc + "['{}_levels']".format(settings['contour_quantity']),
            'Contour levels',
            default=default_contour_levels,
            postcommand=update_figure,
        )

        OMFITx.Button('Set up contours relative to reference...', lambda: that.run(section='efitviewer_contour_picker', out_loc=out_loc))

        if compare_version(omas.__version__, omas_features['sf']) >= 0:
            OMFITx.Entry(
                settings_loc + "['contour_resample_factor']",
                'Resample factor',
                default=3,
                help='Resampling 2d data to higher resolution makes smoother contours.',
                postcommand=update_figure,
            )

    OMFITx.Tab('Export')  # ---------------------------------------------------------------------------------------
    if True:
        OMFITx.Entry(
            settings_loc + "['export_script_filename']",
            'Filename for plot script',
            default="saved_eq_cx_plot_{}_{}".format(device, shot),
            help='The extension will be added.',
        )
        OMFITx.Button('Save plot commands to script', dump_plot_command)

        settings.setdefault('command_box_number', 'new tab')
        OMFITx.ComboBox(
            settings_loc + "['command_box_number']",
            ['new tab'] + list(range(1, 1 + len(OMFITaux['GUI'].command))),
            'Save plot commands to command box #',
            postcommand=box_plot_command,
        )

    OMFITx.Tab('Troubleshoot')  # ---------------------------------------------------------------------------------
    if True:
        OMFITx.CompoundGUI(this, '', section='efitviewer_troubleshooting')

    OMFITx.Tab('Advanced')  # -------------------------------------------------------------------------------------
    if True:
        OMFITx.CompoundGUI(this, '', section='efitviewer_advanced')

    return


def efitviewer_troubleshooting():
    """Provides a GUI for managing troubleshooting options"""
    OMFITx.Button(
        'Reset SSH tunnels, database connections',
        OMFIT.reset_connections,
        help='If the ODS is missing data, try resetting connections and then reloading it.',
    )
    return


def efv_so_general():
    """
    Provides a sub-GUI for handling EFitViewer Special Overlays (efv-so) that are general to all/most devices
    """
    # Custom script
    with OMFITx.same_row():
        OMFITx.CheckBox(
            settings_loc + "['special_systems']['custom_script']",
            'custom_script',
            default=False,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )

        OMFITx.Button('Configure custom script overlay...', lambda: that.run(section='co_efv_custom_script', out_loc=out_loc))

    # Scaled boundary
    with OMFITx.same_row():
        OMFITx.CheckBox(
            settings_loc + "['special_systems']['scaled_boundary']",
            'scaled_boundary',
            default=False,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )

        OMFITx.Button('Configure scaled boundary overlay...', lambda: that.run(section='co_efv_scaled_boundary', out_loc=out_loc))

    # Alternative limiter
    with OMFITx.same_row():
        OMFITx.CheckBox(
            settings_loc + "['special_systems']['alt_limiter']",
            'alt_limiter',
            default=False,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )

        OMFITx.Button('Configure alternative limiter overlay...', lambda: that.run(section='co_efv_alt_limiter', out_loc=out_loc))
    return


def efv_so_d3d():
    """
    Provides a sub-GUI for controlling EFitViewer Special Overlays (EFV-SO) that are specific to the DIII-D device
    """

    # BES is not in the IMAS schema yet as far as I can tell, so it goes under special
    with OMFITx.same_row():
        OMFITx.CheckBox(
            settings_loc + "['special_systems']['beam_emission_spectroscopy']",
            'beam_emission_spectroscopy',
            default=False,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )

        OMFITx.Button(
            'Customize...', lambda: that.run(hw_sys='beam_emission_spectroscopy', section='co_efv_generic', out_loc=out_loc), width=5
        )
    return


def efitviewer_profile_setup():
    """Adds controls for setting up profile displays"""
    OMFITx.Button("Load default profiles setup", setup_default_profiles, updateGUI=True)
    with OMFITx.same_row():
        OMFITx.CheckBox(
            f"{settings_loc}['show_cx']",
            'Show cross-section',
            default=True,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )
        OMFITx.Entry(
            f"{settings_loc}['profiles']['num_profiles']",
            'Number of profiles',
            default=0,
            check=is_int,
            updateGUI=True,
            width=2,
        )
        OMFITx.Entry(
            f"{settings_loc}['profiles']['num_profile_cols']",
            'Number of profile columns',
            default=0,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
            width=2,
        )
    OMFITx.Entry(
        f"{settings_loc}['profiles']['gridspec_kw']",
        'Keywords to pass to gridspec',
        default={},
        help="See gridspec documentation for details. width_ratios is especially useful here, as in {'width_ratios': [2, 1, 1]}",
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    nprof = settings['profiles']['num_profiles']
    if nprof > 0:
        cq_opts, dcq = find_cq_options()
        yopts, ydefault = find_profile_y_options(index)
        with OMFITx.same_row():
            OMFITx.ComboBox(
                f"{settings_loc}['profiles']['sharex']",
                ['none', 'col', 'row', 'all'],
                "Share X axes",
                default='all',
                postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
            )
            OMFITx.ComboBox(
                f"{settings_loc}['profiles']['sharey']",
                ['none', 'col', 'row', 'all'],
                "Share Y axes",
                default='none',
                postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
            )
        OMFITx.CheckBox(
            f"{settings_loc}['profiles']['last_col_right_label']",
            'Put Y-axis labels on the right for the last column',
            default=True,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )
        OMFITx.ComboBox(
            f"{settings_loc}['profiles']['xaxis_quantity']",
            cq_opts + ['r_outboard'],
            "X axis quantity",
            default=dcq,
            state='normal',
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )
        # Plan out arrangement of y axis selectors
        ncol = settings['profiles']['num_profile_cols']
        if ncol == 0:
            ncol = 1
        num_rows = int(np.max([np.ceil(nprof / float(ncol)), 1]))
        profiles_scratch = np.empty([num_rows, ncol], bool)
        profiles_scratch[:] = False
        for i in range(nprof):
            row = i % num_rows
            col = i // num_rows
            profiles_scratch[row, col] = True
        # Make Y axis quantity selectors
        for j in range(ncol):
            OMFITx.Tab(f'Prof col {j}')
            for i in range(num_rows):
                number = i * ncol + j
                if profiles_scratch[i, j]:
                    OMFITx.ComboBox(
                        f"{settings_loc}['profiles']['yaxis_quantity_{number}']",
                        yopts,
                        f'Y-axis quantity for plot {number}, row {i}, col {j}',
                        default=ydefault[number],
                        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                    )


# Call the specified section
printd('Launch efitviewer_gui for section = {}'.format(section), topic='efitviewer')
eval(section)()



import os as _os
from os.path import *


def join(a, *p):
    return '/'.join([a] + list(p))


join.__doc__ = _os.path.join.__doc__


for _item in ['normpath', 'abspath']:
    _str = f'''
def {_item}(path):
    return _os.path.{_item}(path).replace('\\\\', '/')

{_item}.__doc__ = _os.path.{_item}.__doc__
'''
    exec(_str)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_data import Dataset, DataArray, OMFITncDataset
from omfit_classes.utils_math import interp1e

from types import MethodType
from io import StringIO
import numpy as np
from scipy.io import FortranFile

__all__ = ['OMFITGPECbin', 'OMFITGPECascii', 'OMFITGPECnc', 'OMFITldpi']

########################################################### NetCDF files


def lock_items(f):
    """Decorator that should lock up the dictionary-like class after the method is called"""

    @functools.wraps(f)
    def lock(self, *args, **kwargs):
        """Calls the function then modified __setitem__ and __getitem__ to prevent editing keys/values."""
        # first call the function (like load)
        result = f(self, *args, **kwargs)
        # now that we have a good tree representation of the file, don't let anyone edit it
        def locked(self, *args, **kwargs):
            """This is over riden so GPEC outputs cannot be modified directly"""
            raise AssertionError("GPEC output cannot be modified in OMFIT. \n" + "Use to_dataset to get an in-memory copy of the data")

        self.__setitem__ = MethodType(locked, self)

        def jail_break(obj, key):
            """This should copy all the items when they are requested so they cannot be modified in place"""
            return copy.deepcopy(dict.__getitem__(obj, key))

        self.__getitem__ = MethodType(jail_break, self)

        return result

    return lock


class OMFITGPECnc(OMFITncDataset):
    """
    Child of OMFITncDataset, which is a hybrid xarray Dataset and OMFIT SortedDict object.
    This one updates the GPEC naming conventions when it is loaded, and locks the data
    by default so users don't change the code outputs.

    """

    def __init__(self, filename, lock=True, update_naming_conventions=True, export_kw={}, **kwargs):
        self._update_names = update_naming_conventions
        OMFITncDataset.__init__(self, filename, lock=lock, export_kw=export_kw, **kwargs)

    @dynaLoad
    def load(self):
        """
        Loads the netcdf file using the omfit_data importDataset function,
        then distributes those keys and attributes to this class.

        :return:
        """
        OMFITncDataset.load(self)

        # this overrides the lock
        lock = self.lock
        self.set_lock(False)
        if self._update_names:
            try:
                self = self._update_naming_conventions()
            except ValueError as err:
                printe(err)
            self._update_names = False  # stop it from re-updating
        self.set_lock(lock)

    def _update_naming_conventions(self, version=None):
        """

        Update the naming conventions in a dataset to the latest conventions.

        Use this to avoid lots of if statements in cross-checking versions.

        :param dataset:
        :param version: str. Override version attribute.
        :return:

        """
        # set to operate on
        newset = copy.deepcopy(self)

        # version number
        if version is None:
            version = self.attrs.get('version', None)
            if version is None:
                raise ValueError('Must specify version')
        if 'version' in version:
            version = version.split()[-1]  # just the major.minor.patch numbers
        else:
            version = version[1:].split('-')[0]  # newer git tag version label
        version = np.sum(np.array([1, 0.1, 0.001]) * list(map(np.float, version.split('.'))))  # float representation

        translator = OrderedDict()
        if version < 0.3:
            # profile output changes
            translator['xi_psi1'] = 'xigradpsi_dpsi'
            translator['xi_psi'] = 'xigradpsi'
            translator['xi_alpha'] = 'xigradalpha'
        if version < 0.5:
            # control output changes
            translator['b_xn'] = 'b_n_x_fun'
            translator['b_n'] = 'b_n_fun'
            translator['xi_xn'] = 'xi_n_x_fun'
            translator['xi_n'] = 'xi_n_fun'
            translator['dphi'] = 'delta_phi'
            translator['b_xnm'] = 'b_n_x'
            translator['b_nm'] = 'b_n'
            translator['xi_xnm'] = 'xi_n_x'
            translator['xi_nm'] = 'xi_n'
            translator['Phi_X'] = 'Phi_x'
            translator['Phi_EX'] = 'Phi_xe'
            translator['Phi_T'] = 'Phi'
            translator['Phi_ET'] = 'Phi_e'
            translator['X_EVT'] = 'X_eigenvalue'
            translator['X_EDT'] = 'X_eigenvector'
            translator['W_EVX'] = 'W_xe_eigenvalue'
            translator['R_EVX'] = 'R_xe_eigenvalue'
            translator['P_EVX'] = 'P_xe_eigenvalue'
            translator['C_EVX'] = 'C_xe_eigenvalue'
            translator['W_EDX'] = 'W_xe_eigenvector'
            translator['R_EDX'] = 'R_xe_eigenvector'
            translator['P_EDX'] = 'P_xe_eigenvector'
            translator['C_EDX'] = 'C_xe_eigenvector'
            translator['W_EVX_energyv'] = 'W_xe_energyv'
            translator['W_EVX_energys'] = 'W_xe_energys'
            translator['W_EVX_energyp'] = 'W_xe_energyp'
            translator['R_EVX_energyv'] = 'R_xe_energyv'
            translator['R_EVX_energys'] = 'R_xe_energys'
            translator['R_EVX_energyp'] = 'R_xe_energyp'
            translator['W_EVX_A'] = 'W_xe_amp'
            translator['R_EVX_RL'] = 'R_xe_RL'
            translator['O_XT'] = 'O_Xxi_n'
            translator['O_WX'] = 'O_WPhi_xe'
            translator['O_PX'] = 'O_PPhi_xe'
            translator['O_RX'] = 'O_RPhi_xe'
            # profile output changes
            translator['derxi_m_contrapsi'] = 'xigradpsi_dpsi'
            translator['xi_m_contrapsi'] = 'xigradpsi'
            translator['xi_m_contraalpha'] = 'xigradalpha'
            # cylindrical output changes
            translator['b_r_plas'] = 'b_r_plasma'
            translator['b_z_plas'] = 'b_z_plasma'
            translator['b_t_plas'] = 'b_t_plasma'
        if version < 1.4:
            # profile and control netcdf coordinate name changes
            translator['psi_n_rat'] = 'psi_n_rational'
            translator['q_rat'] = 'q_rational'
            translator['m_rat'] = 'm_rational'
            translator['dqdpsi_n_rat'] = 'dqdpsi_n_rational'
        # do this in an explicit loop because some names already exist and need to get replaced in order
        for okey, nkey in list(translator.items()):
            if okey in newset:
                newset = newset.rename({okey: nkey})

        return newset


########################################################### ASCII files

# better label recognition using genfromtxt
for c in '^|<>/':
    if c in np.lib._iotools.NameValidator.defaultdeletechars:
        np.lib._iotools.NameValidator.defaultdeletechars.remove(c)


def structured_array_to_dataset(fromtxt, dims=(), excludedims=(), attrs={}):
    """
    Takes structured array from np.genfromtxt, and creates Dataset
    with dependent/independent data going to dims/data_vars.

    :param fromtxt: structured array.
    :param dims: list. Use these as dimension variable keys.
    :param excludedims: list. Do not use these as dimension variable keys.
    :param attrs: dict. Dataset attributes.

    """
    dataset = Dataset()
    names = list(fromtxt.dtype.names)
    l = len(fromtxt)

    # Set independent/dependent variables (max 3D,min 3pt grid)
    potxnames = []
    for name in names:
        if name not in excludedims:
            potxnames.append(name)
        if len(potxnames) >= 3:
            break
    # by default, don't use psi and q as independent dims (but they are often listed next to each other)
    if np.any([k.startswith('psi') for k in potxnames]) and 'q' in potxnames:
        potxnames.remove('q')
    # catch r,z ("fun") descriptions of the control surface
    if set(potxnames) == set(['r', 'z', 'theta']):
        potxnames = ['theta', 'r', 'z']
    if set(potxnames) == set(['R(m)', 'Z(m)', 'theta(rad)']):  # wall_geom.out
        potxnames = ['theta(rad)']
    # catch crit.out, which has insufficient accuracy in psi so it looks repeated
    if len(potxnames) and potxnames[0] == 'is':
        potxnames = potxnames[:1]
    if not np.all([name in names for name in dims]):
        print('WARNING: Requested axes not available. Using default left column(s).')
    if dims and np.all([name in names for name in dims]):
        nd = len(dims)
        potxnames = list(dims)
    else:
        nd = 1
        npot = len(potxnames)
        if npot > 1 and len(set(fromtxt[potxnames[1]])) < int(l // 2):
            nd += 1
            if npot > 2 and len(set(fromtxt[potxnames[2]])) < int(l // 3):
                nd += 1
    xnames = potxnames[:nd]
    xcols = np.array([np.array(fromtxt[key]).ravel() for key in xnames])
    datashape = []
    for xkey, xcol in zip(xnames, xcols):
        xset = np.sort(list(set(xcol)))  # ordered set. Presumably the table column repeated values if nd>1
        datashape.append(len(xset))
        dataset[xkey] = DataArray(xset, coords={xkey: xset}, dims=(xkey,))
    if np.product(datashape) != l:  # had repeats in dimension variables! Probably psi to closely packed
        # fall back to a simple indexing scheme
        dataset = Dataset()
        xkey = 'row_index'
        xset = np.arange(l)
        dataset[xkey] = DataArray(xset, coords={xkey: xset}, dims=(xkey,))
        xnames = ['row_index']
        nd = 1
        datashape = [l]
        # raise ValueError(
        #     "The table rows are not fully described by the unique valued dimensions {:}".format(potxnames[:nd]))

    # set the data_vars, allowing for real and imaginary columns
    ynames = [name for name in names if name not in xnames]
    for name in ynames:
        if 'real' in name:
            newname = name.replace('real', '')
            newname = newname.lstrip('(').rstrip(')')
            if not newname:  # the columns just say "real    imag"
                newname = "var_0"
            y = np.reshape(fromtxt[name] + 1j * fromtxt[name.replace('real', 'imag')], datashape)
        elif not 'imag' in name:
            newname = name
            y = np.reshape(fromtxt[name], datashape)
        dataset[newname] = DataArray(y, coords=dataset.coords, dims=xnames)
    dataset.attrs.update(attrs)

    return dataset


def text_to_attrs(txt):
    """Parse text with variables defined using 'name = value' syntax."""
    preamble = txt.split()
    params = []
    names = []
    while preamble.count('='):  # parameter specifications
        idx = preamble.index('=')
        param = preamble[idx + 1]
        name = preamble[idx - 1]
        name.translate({ord(char): '' for char in r'()[]{}\/*-+^%.,:!@#&'})
        if name in names and idx > 1:
            name = ' '.join(preamble[idx - 2 : idx])
            name = name.translate({ord(char): '' for char in r'()[]{}\/*-+^%.,:!@#&'})
        elif name in names:
            for sfx in range(1, 100):
                if name + str(sfx) not in names:
                    name += str(sfx)
                    break
        try:
            params.append((name, float(param.rstrip('.,'))))  # if its a number
        except Exception:
            params.append((name, param))  # if its not a number
        names.append(name)
        preamble.remove(preamble[idx - 1])
        preamble.remove(str(param))
        preamble.remove('=')
    return dict(params)


def ascii_to_dataset(fname, squeeze=False, dims=(), excludedims=('l',), coords=(), maxnumber=999, verbose=False):
    """
    Get the data from any gpec output as a list of python
    class-type objects using np.genfromtxt.

    :param fname: str. Path of gpec output file.
    :param squeeze: bool. True uses the update method to form a single data object from multiple tables.
        str. Concatenate all tables along the specified dimension or attribute.
    :param dims: list. Use these as the independent data.
    :param excludedims: list. Do not use these as the independent data.
    :param maxnumber: int. Reads only first maxnumber of tables.
    :param verbose: bool. Print verbose status messages.

    :returns: list. Class objects for each block of data in the file.

    """
    collection = []
    dcollection = []
    pcollection = []

    with open(fname) as f:
        # Get preamble
        preamble = ''
        lastline = ''
        for lstart, line in enumerate(f):
            try:
                test = float(line.split()[0])
                break
            except Exception:
                preamble += lastline
                lastline = line
        f.seek(0)
        try:  # Simple file : single table
            # data = np.genfromtxt(f,skip_header=lstart-1,names=True,dtype=np.float64)
            raise ValueError  # genfromtxt read pfile (multiple 2-column tables) as one... '\r' bug
            pcollection.append(preamble)
            dcollection.append(data)
        # look through file manually
        except ValueError:
            f.seek(0)
            # clean up messy dcon output formats
            if 'dcon.out' in fname or 'stride.out' in fname:
                lines = f.read()
                lines = lines.replace('mu0 p', 'mu0p')
                lines = lines.replace(' abs ', ' abs')
                lines = lines.replace(' re ', ' real')
                lines = lines.replace(' im ', ' imag')
                lines = lines.replace('*', '')
                lines = lines.replace('q          q1', 'qs        qs1')
                lines = lines.replace('i     re', 'ipert re')
                lines = lines.replace(
                    '   isol  imax   plasma     vacuum     total\n', '   isol  imax   plasma     vacuum realtotal imagtotal\n'
                )
                lines = StringIO(lines).readlines()
            else:
                lines = f.readlines()
            top, bot = 0, 0
            count = 0
            length = len(lines)
            while bot < length and top < (length - 2) and count < maxnumber:
                preamble = ''
                lastline = ''
                for i, line in enumerate(lines[bot:]):
                    try:
                        # Find top defined as start of numbers
                        test = float(line.split()[0])
                        top = bot + i
                        # Throw out single lines of numbers
                        # if not lines[top+1].translate({ord(char): '' for char in ' \n\t\r'}):#=='\n':
                        #    raise ValueError
                        # Find bottom defined by end of numbers
                        for j, bline in enumerate(lines[top:]):
                            try:
                                test = float(bline.split()[0])
                            except Exception:
                                break  # end of numbers
                        # Throw out single lines of numbers
                        if j == 1:
                            # but include them as preamble (DCON one-liners)
                            vals = lines[top]
                            keys = lines[top - 1]
                            if not keys.translate({ord(char): '' for char in ' \n\t'}):  # empty line
                                keys = lines[top - 2]
                            if '=' not in keys and len(keys.split()) == len(vals.split()):
                                for k, v in zip(keys.split(), vals.split()):
                                    preamble += '{:} = {:}\n'.format(k, v)
                            raise ValueError
                        else:
                            bot = top + j + 1
                        break
                    except Exception:
                        preamble += lastline
                        lastline = line
                if line == lines[-1] and line == lastline:
                    break  # end of file without another table

                # include headers
                top -= 1
                if not lines[top].translate({ord(char): '' for char in ' \n\t'}):  # empty line
                    top -= 1
                skipfoot = length - bot
                f.seek(0)
                table = lines[top:bot]
                if '\n' in table:  # empty space
                    table.remove('\n')
                data = np.genfromtxt(StringIO(''.join(table)), names=True, deletechars='?', dtype=np.float64)
                pcollection.append(preamble)
                dcollection.append(data)
                count += 1
        # turn arrays into classes
        for i, (data, preamble) in enumerate(zip(dcollection, pcollection)):
            if verbose:
                printi("Casting table " + str(i + 1) + " into Dataset.")
            attrs = text_to_attrs(preamble)
            ds = structured_array_to_dataset(data, attrs=attrs, dims=dims, excludedims=excludedims)
            try:
                ds = structured_array_to_dataset(data, attrs=attrs, dims=dims, excludedims=excludedims)
            except ValueError as err:
                # sometimes diagnostics of splines used repeating ix values in the first column
                # this is confusing when sorting out the dims... try ignoring that column when looking for dims
                printe(str(err))
                baddim = str(err).split("'")[-2]
                excludedims = list(excludedims) + [baddim]
                ds = structured_array_to_dataset(data, attrs=attrs, dims=dims, excludedims=excludedims)
            ds = ds.set_coords(set(coords).intersection(set(ds.data_vars.keys())))

            collection.append(ds)

    # force all to a single object
    if squeeze:
        if isinstance(squeeze, bool):
            # choose some default squeeze dimensions
            if np.any(['isol' in ds.attrs for ds in collection]):
                squeeze = 'isol'  # common in DCON
            if np.any(['ia' in ds.attrs for ds in collection]):
                squeeze = 'ia'  # used in EQUIL 2d.out
            else:
                if np.any(['mode' in ds.attrs for ds in collection]):
                    squeeze = 'mode'  # common in GPEC
        if isinstance(squeeze, str):
            # This grouping catches cases where one dim of 2D data was split into separate tables,
            # but 1D data for that dim also exists
            index = 0
            lc = len(collection)
            dsq = Dataset()
            for i in range(lc):
                if index >= lc:
                    break
                series2d = []
                for j in range(index, lc):
                    if np.all(list(collection[j].keys()) == list(collection[index].keys())):
                        series2d.append(collection[j])
                        index = j
                index += 1
                # combine the series of similar blocks
                ds2d = combine_datasets(series2d, squeeze)
                # then safely combine that with the rest of the data
                dsq = combine_datasets((dsq, ds2d), squeeze)
        else:
            try:
                dsq = combine_datasets(collection)
            except Exception as err:
                printe("Unable to merge multiple tables found in file, returning only the last one")
                dsq = collection[-1]
        return dsq

    return collection


def combine_datasets(datasets, dim=''):
    """
    Combine multiple datasets, allowing for the concatenating dimension to have been
    written in the headers of multiple tables (essentially writing multi-dimensional data
    as a series of 1D tables) and thus appearing in the dataset attributes.

    :param datasets: list. Datasets to be combined into one.
    :param dim: str. Dimension along which to concatenate. Can be a common attribute.
    :return:

    """
    new = Dataset()
    for i, d in enumerate(datasets):
        # allow a new dimension specified in each tables preamble
        if dim in d.attrs:
            x = np.atleast_1d(d.attrs.pop(dim))
            d *= DataArray(np.ones_like(x), coords=[(dim, x)])
        # concat along an existing dimension if necessary
        if i == 0:
            new.update(d)
        elif dim in d.dims and dim in new.dims and not (np.all(d[dim] == new[dim]) and len(d[dim] == new[dim]) > 0):
            new = xarray.concat((new, d), dim, data_vars='minimal')
        else:
            new.update(d)
        new.attrs.update(d.attrs)
    return new


class OMFITGPECascii(OMFITncDataset, OMFITascii):
    """
    This class parses ascii tables in a very general way that catches most of the oddities
    in GPEC output file formatting (mostly inherited from DCON).

    The data is stored as a Dataset in memory but has no save method, so the original
    file is never modified.

    """

    def __init__(self, filename, lock=True, export_kw={}, **kwargs):
        OMFITascii.__init__(self, filename, **kwargs)
        OMFITncDataset.__init__(self, filename, lock=lock, export_kw=export_kw)

    @dynaLoad
    def load(self):
        """
        Load ascii tables into memory.
        """
        try:
            # try to squeeze along the isol dimension, which DCON likes to split into separate tables
            data = ascii_to_dataset(self.filename, squeeze=True)
            # try to put this into a OMFITncDataset type object-Dataset hybrid
            lock = self.lock
            self.set_lock(False)
            self.update(data)
            self.attrs = data.attrs
            self.set_lock(lock)
        except Exception as err:
            printe("WARNING: Could not parse " + self.filename.split('/')[-1])
            # printe(repr(err))
            raise err

        self.dynaLoad = False

    @dynaSave
    def save(self):
        """
        This method can detect if .filename was changed and if so, makes a copy from the original .filename
        (saved in the .link attribute) to the new .filename
        """
        OMFITobject.save(self)


########################################################### Binary Files


def _get_binary_mapping(gpec_home='$GPECHOME'):
    """Get a mapping of all the binary variables that the xdraw application knows"""
    import json

    # list all the files in the draw directory
    files = []
    for k in os.listdir(gpec_home + '/draw'):
        pth = os.path.abspath(gpec_home + '/draw/' + k)
        print(pth)
        if os.path.isfile(pth) and k.endswith('.in'):
            files.append(pth)
    bin_map = {}
    for pth in files:
        with open(pth, 'r') as f:
            lines = f.readlines()
        # find markers
        iname, idstart, idend, iistart, iiend = -1, -1, -1, -1, -1
        for i, l in enumerate(lines):
            if 'filename' in l:
                iname = i + 1
            if l.startswith('independent variable names'):
                iistart = i + 1
            if l.startswith('\n') and iistart > 0 and iiend < 0:
                iiend = i
            if l.startswith('dependent variable names') or l.startswith('variable names'):
                idstart = i + 1
            if l.startswith('\n') and idstart > 0 and idend < 0:
                idend = i
        # parse variable names
        name = lines[iname].split('.')[0]
        vars = ['_'.join(l.split()[1:]).lower() for l in lines[idstart:idend]]
        dims = ['_'.join(l.split()[1:]).lower() for l in lines[iistart:iiend]]
        type_code = lines[0].split()[-1]
        if type_code == 'G':  # 1D "graphs"
            dims = [vars.pop(0)]
        # store in a master map
        bin_map[name] = dict(vars=vars, dims=dims, type=type_code)

    print(json.dumps(bin_map, indent=4))
    return bin_map


bin_mapping = {
    "deltap": {"dims": ["r"], "type": "G", "vars": ["q", "psi", "psi1", "u1", "u2", "log10_ca1", "log10_ca2", "log10_|z|"]},
    "ahb": {
        "dims": ["theta"],
        "type": "G",
        "vars": [
            "re_bn",
            "re_bp",
            "re_bt",
            "im_bn",
            "im_bp",
            "im_bt",
            "theta_hamada",
            "theta_pest",
            "theta_equal_arc",
            "theta_boozer",
            "dtheta_hamada",
            "dtheta_pest",
            "dtheta_equal_arc",
            "dtheta_boozer",
        ],
    },
    "chord": {"dims": ["r"], "type": "G", "vars": ["q", "xi_psi", "b_psi", "xi_norm", "b_norm"]},
    "kwmats": {
        "dims": ["psifac"],
        "type": "G",
        "vars": [
            "real_ak",
            "imag_ak",
            "real_bk",
            "imag_bk",
            "real_ck",
            "imag_ck",
            "real_dk",
            "imag_dk",
            "real_ek",
            "imag_ek",
            "real_hk",
            "imag_hk",
        ],
    },
    "ktmats": {
        "dims": ["psifac"],
        "type": "G",
        "vars": [
            "real_ak",
            "imag_ak",
            "real_bk",
            "imag_bk",
            "real_ck",
            "imag_ck",
            "real_dk",
            "imag_dk",
            "real_ek",
            "imag_ek",
            "real_hk",
            "imag_hk",
        ],
    },
    "xbnormal": {
        "dims": ["psifac"],
        "type": "G",
        "vars": ["real_xi", "imag_xi", "real_b", "imag_b", "real_jac.b.delpsi", "imag_jac.b.delpsi"],
    },
    "metric": {
        "dims": ["psi"],
        "type": "G",
        "vars": [
            "log10_psi",
            "log10_1-psi",
            "log10_ffco1",
            "arg_ffco1",
            "log10_ffco2",
            "arg_ffco2",
            "log10_ffco3",
            "arg_ffco3",
            "log10_ffco4",
            "arg_ffco4",
            "log10_ffco5",
            "arg_ffco5",
            "log10_ffco6",
            "arg_ffco6",
        ],
    },
    "sum2": {"dims": ["q0"], "type": "G", "vars": ["inval", "psifac", "di", "deltap"]},
    "surface": {
        "dims": ["theta"],
        "type": "G",
        "vars": [
            "re_tn",
            "re_tp",
            "re_tt",
            "im_tn",
            "im_tp",
            "im_tt",
            "re_bn",
            "re_bp",
            "re_bt",
            "im_bn",
            "im_bp",
            "im_bt",
            "theta_hamada",
            "theta_pest",
            "theta_equal_arc",
            "theta_boozer",
            "dtheta_hamada",
            "dtheta_pest",
            "dtheta_equal_arc",
            "dtheta_boozer",
        ],
    },
    "metyx": {"dims": ["theta"], "type": "G", "vars": ["j11", "j12", "j13", "j21", "j22", "j23", "j33"]},
    "sol10": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "solutions": {"dims": ["psifac"], "type": "G", "vars": ["rho", "q", "real_xi", "imag_xi", "real_b", "imag_b"]},
    "crit": {"dims": ["psifac"], "type": "G", "vars": ["log_10_psifac", "log_10_(m_-_n_q)", "q", "crit"]},
    "sol+00": {"dims": ["psifac"], "type": "G", "vars": ["logpsi1", "logpsi2", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "ode": {"dims": ["t"], "type": "G", "vars": ["psi", "theta", "zeta", "p_psi", "p_theta", "p_zeta"]},
    "bal1": {"dims": ["theta"], "type": "G", "vars": ["dbdb0", "dbdb1", "dbdb2", "kapw0", "kapw1"]},
    "fixup": {"dims": ["psifac"], "type": "G", "vars": ["log10_psifac", "log10_uration"]},
    ";results/time193/0": {
        "dims": ["theta"],
        "type": "G",
        "vars": [
            "re_bn",
            "re_bp",
            "re_bt",
            "im_bn",
            "im_bp",
            "im_bt",
            "theta_hamada",
            "theta_pest",
            "theta_equal_arc",
            "theta_boozer",
            "dtheta_hamada",
            "dtheta_pest",
            "dtheta_equal_arc",
            "dtheta_boozer",
        ],
    },
    "rzphi": {
        "dims": ["psi"],
        "type": "G",
        "vars": ["log10_psi", "log10_1-psi", "log10_r2", "arg_r2", "log10_eta", "arg_eta", "log10_phi", "arg_phi"],
    },
    "xbtangent": {"dims": ["psifac"], "type": "G", "vars": ["real_xi", "imag_xi", "real_b", "imag_b"]},
    "sol11": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "sol12": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "avec": {"dims": ["psifac"], "type": "G", "vars": ["-phi", "chi", "1/v1^2"]},
    ";results/trial/1": {
        "dims": ["theta"],
        "type": "G",
        "vars": ["re_bnorm", "im_bnorm", "re_xinorm", "im_xinorm", "theta_hamada", "theta_pest", "theta_equal_arc", "theta_boozer"],
    },
    "results/time129v2fixb/0": {
        "dims": ["psifac"],
        "type": "G",
        "vars": ["rho", "f", "mu0_p", "q", "arcsinh_di", "arcsinh_dr", "arcsinh_h", "arcsinh_ca1"],
    },
    "dcon": {"dims": ["psifac"], "type": "G", "vars": ["rho", "f", "mu0_p", "q", "arcsinh_di", "arcsinh_dr", "arcsinh_h", "arcsinh_ca1"]},
    "rz": {"dims": ["theta"], "type": "G", "vars": ["r2", "eta"]},
    "contour": {"dims": ["time", "r", "z"], "type": "M", "vars": ["xi", "b"]},
    "frs": {"dims": ["q0"], "type": "G", "vars": ["rs", "deltap"]},
    "pmodb_2d": {
        "dims": ["time", "r", "z"],
        "type": "M",
        "vars": ["eulerian_re_mod_b", "eulerian_im_mod_b" "lagrangian_re_mod_b", "lagrangian_im_mod_b"],
    },
    "psi_in": {"dims": ["time", "r", "z"], "type": "M", "vars": ["psi"]},
    "prof": {"dims": ["psifac"], "type": "G", "vars": ["rho", "f", "mu0_p", "q"]},
    "fourfit": {
        "dims": ["psi"],
        "type": "G",
        "vars": ["amat_new", "bmat_new", "cmat_new", "dmat_new", "emat_new", "hmat_new", "fmat_new", "kmat_new", "gmat_new"],
    },
    "cyl": {"dims": ["beta0"], "type": "G", "vars": ["q0", "psis", "di", "deltap"]},
    "xbnormal_2d": {"dims": ["time", "r", "z"], "type": "M", "vars": ["re_xi_normal", "im_xi_normal"]},
    "sq_in": {"dims": ["psi"], "type": "G", "vars": ["f", "p", "q", "rho"]},
    "gsec": {"dims": ["time", "r", "z"], "type": "M", "vars": ["flux1", "flux2", "source", "total", "error", "errlog"]},
    "rz_in": {"dims": ["theta"], "type": "G", "vars": ["r", "z"]},
    "fmat": {
        "dims": ["psi"],
        "type": "G",
        "vars": [
            "log_10_psi",
            "log_10_1-psi",
            "log_10_fmat01",
            "arg_fmat01",
            "log_10_fmat02",
            "arg_fmat02",
            "log_10_fmat03",
            "arg_fmat03",
            "log_10_fmat04",
            "arg_fmat04",
            "log_10_fmat05",
            "arg_fmat05",
            "log_10_fmat06",
            "arg_fmat06",
            "log_10_fmat07",
            "arg_fmat07",
            "log_10_fmat08",
            "arg_fmat08",
            "log_10_fmat09",
            "arg_fmat09",
            "log_10_fmat10",
            "arg_fmat10",
            "log_10_fmat11",
            "arg_fmat11",
            "log_10_fmat12",
            "arg_fmat12",
            "log_10_fmat13",
            "arg_fmat13",
        ],
    },
    "pflux_im_2d": {"dims": ["time", "r", "z"], "type": "M", "vars": ["im_perturbed_flux"]},
    "gsei": {"dims": ["psi"], "type": "G", "vars": ["term1", "term2", "totali", "errori", "log10_errori"]},
    "vbnormal": {"dims": ["psifac"], "type": "G", "vars": ["real_jac.b.delpsi", "imag_jac.b.delpsi"]},
    "gmat": {
        "dims": ["psi"],
        "type": "G",
        "vars": [
            "log_10_psi",
            "log_10_1-psi",
            "log_10_gmat01",
            "arg_gmat01",
            "log_10_gmat02",
            "arg_gmat02",
            "log_10_gmat03",
            "arg_gmat03",
            "log_10_gmat04",
            "arg_gmat04",
            "log_10_gmat05",
            "arg_gmat05",
            "log_10_gmat06",
            "arg_gmat06",
            "log_10_gmat07",
            "arg_gmat07",
            "log_10_gmat08",
            "arg_gmat08",
            "log_10_gmat09",
            "arg_gmat09",
            "log_10_gmat10",
            "arg_gmat10",
            "log_10_gmat11",
            "arg_gmat11",
            "log_10_gmat12",
            "arg_gmat12",
            "log_10_gmat13",
            "arg_gmat13",
        ],
    },
    "vbnormal_spectrum": {"dims": ["time", "poloidal_harmonics", "psi"], "type": "M", "vars": ["jac.b.delpsi"]},
    "debug": {"dims": ["psi"], "type": "G", "vars": ["theta", "r", "z", "xi_psi", "b_psi", "xi_norm", "b_norm"]},
    "xbtangent_2d": {"dims": ["time", "r", "z"], "type": "M", "vars": ["re_xi_tangent", "im_xi_tangent"]},
    "bnormal_spectrum": {"dims": ["time", "poloidal_harmonics", "psi"], "type": "M", "vars": ["jac.b.delpsi"]},
    "flint": {"dims": ["eta"], "type": "G", "vars": ["y1", "y2", "r", "z", "err"]},
    "metxy": {"dims": ["psi"], "type": "G", "vars": ["j11", "j12", "j13", "j21", "j22", "j23", "j33"]},
    "bal2": {"dims": ["theta"], "type": "G", "vars": ["asinh_y1", "asinh_y2", "y2/dy1", "dy2/y1", "asinh_ca1", "asinh_ca2"]},
    "plen": {"dims": ["ipert"], "type": "G", "vars": ["psifac", "asinh_eval"]},
    "line": {"dims": ["r"], "type": "G", "vars": ["theta", "z", "xi_psi", "b_psi", "xi_rho", "b_rho", "jstep"]},
    ";results/betaN/3": {
        "dims": ["theta"],
        "type": "G",
        "vars": [
            "re_bn",
            "re_bp",
            "re_bt",
            "im_bn",
            "im_bp",
            "im_bt",
            "theta_hamada",
            "theta_pest",
            "theta_equal_arc",
            "theta_boozer",
            "dtheta_hamada",
            "dtheta_pest",
            "dtheta_equal_arc",
            "dtheta_boozer",
        ],
    },
    "lar": {"dims": ["r"], "type": "G", "vars": ["bp", "bt", "pp", "sigma", "q", "p", "f", "psi"]},
    "kmat": {
        "dims": ["psi"],
        "type": "G",
        "vars": [
            "log_10_psi",
            "log_10_1-psi",
            "log_10_kmat01",
            "arg_kmat01",
            "log_10_kmat02",
            "arg_kmat02",
            "log_10_kmat03",
            "arg_kmat03",
            "log_10_kmat04",
            "arg_kmat04",
            "log_10_kmat05",
            "arg_kmat05",
            "log_10_kmat06",
            "arg_kmat06",
            "log_10_kmat07",
            "arg_kmat07",
            "log_10_kmat08",
            "arg_kmat08",
            "log_10_kmat09",
            "arg_kmat09",
            "log_10_kmat10",
            "arg_kmat10",
            "log_10_kmat11",
            "arg_kmat11",
            "log_10_kmat12",
            "arg_kmat12",
            "log_10_kmat13",
            "arg_kmat13",
        ],
    },
    "lar2": {"dims": ["psi"], "type": "G", "vars": ["rho", "q", "g11", "g22", "g33", "f", "g", "k"]},
    "lar1": {"dims": ["psi"], "type": "G", "vars": ["rho", "q", "g11", "g22", "g33", "f", "g", "k"]},
    "pmodb": {"dims": ["psifac"], "type": "G", "vars": ["real_eulerian", "imag_eulerian", "real_lagrangian", "imag_lagrangian"]},
    "scan": {"dims": ["asinh_re_s"], "type": "G", "vars": ["asinh_im_s", "re_det", "im_det", "asinh_re_det", "asinh_im_det"]},
    "diffw": {"dims": ["m"], "type": "G", "vars": ["re_diffw", "im_diffw", "re_bctheta", "im_bctheta", "re_bczeta", "im_bczeta"]},
    "eval": {"dims": ["ipert"], "type": "G", "vars": ["psifac", "log_10_psifac", "log_10_(m_-_n_q)", "log_10_hu", "q", "eval"]},
    "sol07": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "det": {"dims": ["de"], "type": "G", "vars": ["do", "det", "deta", "error"]},
    "orbit": {
        "dims": ["tau"],
        "type": "G",
        "vars": ["r", "z", "phi", "error", "bmod", "rl", "rlfac", "dmufac", "psi", "rho", "vpar", "x", "y"],
    },
    "ipout_contra": {
        "dims": ["psifac"],
        "type": "G",
        "vars": [
            "rho",
            "q",
            "real_xipsi",
            "imag_xipsi",
            "real_xitheta",
            "imag_xitheta",
            "real_xizeta",
            "imag_xizeta",
            "real_bpsi",
            "imag_bpsi",
            "real_btheta",
            "imag_btheta",
            "real_bzeta",
            "imag_bzeta",
            "real_xis",
            "imag_xis",
            "real_dxipsi",
            "imag_dxipsi",
            "real_dbpsi",
            "imag_dbpsi",
            "real_delta",
            "imag_delta",
            "delta",
        ],
    },
    "2d": {"dims": ["theta"], "type": "G", "vars": ["r2", "deta", "eta", "dphi", "r", "z", "jac"]},
    "sol09": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "sol08": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "gse": {"dims": ["theta"], "type": "G", "vars": ["psi", "flux1", "flux2", "source", "total", "error", "log10_error"]},
    "sum": {
        "dims": ["outval"],
        "type": "G",
        "vars": [
            "inval",
            "psilow",
            "psihigh",
            "amean",
            "rmean",
            "aratio",
            "kappa",
            "delta1",
            "delta2",
            "li1",
            "li2",
            "li3",
            "ro",
            "zo",
            "psio",
            "betap1",
            "betap2",
            "betap3",
            "betat",
            "betan",
            "bt0",
            "q0",
            "qmin",
            "qmax",
            "qa",
            "crnt",
            "asinh_plasma1",
            "asinh_vacuum1",
            "asinh_total1",
        ],
    },
    "sol03": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "sol02": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "sol01": {"dims": ["psifac"], "type": "G", "vars": ["log10_psifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "pflux_re_2d": {"dims": ["time", "r", "z"], "type": "M", "vars": ["re_perturbed_flux"]},
    "root": {"dims": ["log10_|s|"], "type": "G", "vars": ["re_root", "im_root"]},
    "sol06": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "sol05": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
    "sol04": {"dims": ["psifac"], "type": "G", "vars": ["logpsifac", "log10_u1", "log10_u2", "log10_ca1", "log10_ca2"]},
}


class OMFITGPECbin(OMFITncDataset):
    r"""
    OMFIT class used to interface with DCON and GPEC binary files.

    :param filename: String. File path (file name must start with original output name)

    :param \**kw: All additional key word arguments passed to OMFITobject

    """

    def __init__(self, filename, lock=True, export_kw={}, **kwargs):
        OMFITncDataset.__init__(self, filename, lock=lock, export_kw=export_kw, **kwargs)

    @dynaLoad
    def load(self):
        """
        Parse GPEC binaries using the xdraw mapping for variable names
        and expected binary structures.

        """
        data_set = Dataset()
        name = self.filename.split('/')[-1].split('.')[0]

        # warn user if they just opened a big (750KB+) slow file
        if os.path.getsize(self.filename) * 1e-3 > 750:
            print('Loading {:}. This may take a second...'.format(name))

        # find out if we know this file's format/vars
        found_name = False
        for k, v in list(bin_mapping.items()):
            if k == name:
                found_name = True
                vars = bin_mapping[name]['vars']
                dims = bin_mapping[name]['dims']
                dtype = bin_mapping[name]['type']
                break

        # if it is a "graph" file, data is written in column format
        # and breaks signal some sort of new line (often used to plot "slices" of 2D data)
        if found_name and dtype == 'G':
            index = 0
            # read the data if we do recognize it
            with FortranFile(self.filename, 'r') as f:
                data = []
                for i in range(int(1e8)):
                    try:
                        tmp = f.read_reals(dtype='f4')
                    except Exception:
                        break
                    if len(tmp):  # defence against DCON's empty lines
                        data.append(tmp)
                    else:
                        index += 1
                data = np.vstack(data).T
            # this would be nice, but DCON sometimes puts weird breaks in the data :/
            # with open(self.filename, 'r') as f:
            #     data = np.fromfile(f, dtype='float32', count=-1)

            vars = bin_mapping[name]['vars']
            # are there a lot of repeated profiles?
            if np.mod(data.shape[1], index):
                index = 1  # probably broke up profiles into segments for some reason (coloring)
            x = data[0].reshape((-1, index), order='F')
            if index > 1 and np.all(x[:, 0] == x[:, -1]):
                data = data.reshape((data.shape[0], -1, index), order='F')
                dims = (bin_mapping[name]['dims'][0], 'index')
                coords = {dims[0]: x[:, 0], 'index': np.arange(index)}
            else:
                dims = (bin_mapping[name]['dims'][0],)
                coords = {dims[0]: data[0]}
            # load up the data into memory as a Dataset (temporary because we have no save method)
            for i, k in enumerate(vars):
                data_set[k] = DataArray(data[i + 1], coords=coords, dims=dims)

        # if it is a 'M' file, data from each variable is all dumped at once (i.e. row format)
        # header integers specify sizes, then coordinates are dumped together, then variables one at a time
        elif found_name and dtype == 'M':
            # read the data if we do recognize it
            with FortranFile(self.filename, 'r') as f:
                # read header integers
                itime = f.read_ints(dtype=np.int32)
                shape = f.read_ints(dtype=np.int32) + 1
                # read dims
                dim_data = f.read_reals(dtype='f4').reshape(shape[0], shape[1], 2, order='F')
                # read vars
                var_data = []
                for i in range(len(vars)):
                    var_data.append(f.read_reals(dtype='f4').reshape(*shape, order='F'))
                var_data = np.array(var_data)

            # slice the dims, ignoring time
            if np.all(np.all(dim_data[:, :, 0].T == dim_data[:, 0, 0], axis=0)) & np.all(
                np.all(dim_data[:, :, 1] == dim_data[0, :, 1], axis=0)
            ):
                coords = {dims[1]: dim_data[:, 0, 0], dims[2]: dim_data[0, :, 1]}
                dims = dims[1:][::-1]
            else:
                coords = {'dim_0': np.arange(dim_data.shape[0]), 'dim_1': np.arange(dim_data.shape[1])}
                for i, k in enumerate(dims[1::]):
                    data_set[k] = DataArray(dim_data[:, :, i - 1], coords=coords, dims=['dim_0', 'dim_1'])
                dims = ['dim_0', 'dim_1']
            # load up the data into memory as a Dataset (temporary because we have no save method)
            for i, k in enumerate(vars):
                if len(coords[dims[0]]) == var_data[i].shape[0] and len(coords[dims[1]]) == var_data[i].shape[1]:
                    data_set[k] = DataArray(var_data[i], coords=coords, dims=dims)  # definitely true with dim_0, dim_1
                else:
                    data_set[k] = DataArray(var_data[i].T, coords=coords, dims=dims)
        else:
            raise ValueError("Unknown parsing for {:}".format(name))
            # printe("WARNING: Unknown parsing for {:}".format(name))

        lock = self.lock
        self.set_lock(False)
        self.update(data_set)
        self.attrs = {}
        self.set_lock(lock)

        self.dynaLoad = False

        return

    @dynaSave
    def save(self):
        """
        This method can detect if .filename was changed and if so, makes a copy from the original .filename
        (saved in the .link attribute) to the new .filename
        """
        OMFITobject.save(self)


class OMFITldpi(OMFITncDataset):
    r"""
    OMFIT class used to interface with L. Don Pearlstein's inverse equilibrium code.

    :param filename: String. File path (file name must start with original output name)

    :param \**kw: All additional key word arguments passed to OMFITobject

    """

    def __init__(self, filename, lock=True, export_kw={}, **kwargs):
        OMFITncDataset.__init__(self, filename, lock=lock, export_kw=export_kw, **kwargs)

    @dynaLoad
    def load(self):
        """
        Parse ldp_i binaries using the DCON read_eq_ldp_i as a guide for variable names
        and expected binary structures.

        """
        data_set = Dataset()
        name = self.filename.split('/')[-1].split('.')[0]

        # warn user if they just opened a big (750KB+) slow file
        if os.path.getsize(self.filename) * 1e-3 > 750:
            print('Loading {:}. This may take a second...'.format(name))

        # header integers specify sizes, then variables are dumped one at a time
        vars = SortedDict()
        with FortranFile(self.filename, 'r') as binfile:
            # read header integers
            mxmy = binfile.read_ints(dtype=np.int32)
            # read 1d vars
            # Note that in the SPECTOR example, every other number in the vars is garbage
            vars['psi'] = binfile.read_reals(dtype='f8')
            vars['f'] = binfile.read_reals(dtype='f8')
            vars['p'] = binfile.read_reals(dtype='f8')
            vars['q'] = binfile.read_reals(dtype='f8')
            # read 2d vars
            r = binfile.read_reals(dtype='f8').reshape(mxmy[0], mxmy[1], order='C')
            z = binfile.read_reals(dtype='f8').reshape(mxmy[0], mxmy[1], order='C')

        # coordinates
        psio = vars['psi'][-1] - vars['psi'][0]
        psin = (vars['psi'] - vars['psi'][0]) / psio
        theta = np.linspace(-np.pi, np.pi, mxmy[1], endpoint=True)
        # helicity **What is the proper way to get this? From q? From psio? Both?**
        if psio < 0 or np.nanmean(vars['q']) < 0:
            helicity = -1  # is this true? What is the convention?
        else:
            helicity = 1
        # store 1d vars
        coords = {'psi_n': psin}
        dims = ['psi_n']
        for k, v in vars.items():
            data_set[k] = DataArray(v, coords=coords, dims=dims)
        # store 2d vars
        coords = {'psi_n': psin, 'theta': theta}
        dims = ['psi_n', 'theta']
        data_set['R'] = DataArray(r, coords=coords, dims=dims)
        data_set['Z'] = DataArray(z, coords=coords, dims=dims)

        # update the ncDataset in-memory representation of the file. Don't let users mess with it
        lock = self.lock
        self.set_lock(False)
        self.update(data_set)
        self.attrs = {'helicity': helicity}
        self.set_lock(lock)

        self.dynaLoad = False

        return

    @dynaSave
    def save(self):
        """
        This method can detect if .filename was changed and if so, makes a copy from the original .filename
        (saved in the .link attribute) to the new .filename
        """
        OMFITobject.save(self)

    def plot(self, only2D=False, qcontour=False, levels=10, **kwargs):
        """
        Quick plot of LDPI equilibrium.

        :param only2D: bool. Plot only the 2D contour of psi_n (or optionally q)
        :param qcontour: bool. If true, levels of q are contoured in 2D plot (default is psi_n)
        :param levels: int or np.ndarray. Number of (or explicit values of) levels to be contoured in 2D plot. None plots all levels in file
        :param kwargs: All other key word arguments are passed to the matplotlib plot function
        :return:
        """
        # default to all levels in the file
        if levels is None:
            levels = self['psi_n'].values
        # convert q levels into psi_n levels
        elif qcontour:
            # set some number of equally spaced levels
            if is_int(levels):
                levels = np.linspace(self['q'].min(), self['q'].max(), levels + 1, endpoint=True)[1:]  # skip psi_n=0
            levels = interp1e(self['q'], self['psi_n'])(levels)
        # set some number of equally spaced levels
        elif is_int(levels):
            levels = np.linspace(0, 1, levels + 1, endpoint=True)[1:]  # skip psi_n=0

        def plot_2d(ax):
            """
            Kind of like a custom contour using the fact that the r,z arrays trace out flux surfaces.
            Using plot makes the key word arguments consistent with 1D plots too, which is nice.
            :param ax: Axes.
            :return: list of lines
            """
            psi = self['psi_n'].sel(psi_n=levels, method='nearest')
            r = self['R'].sel(psi_n=psi).values
            z = self['Z'].sel(psi_n=psi).values
            x = np.vstack((r.T, np.nan * r[:, 0])).T.ravel()
            y = np.vstack((z.T, np.nan * z[:, 0])).T.ravel()

            l = ax.plot(x, y, **kwargs)
            ax.set_xlabel('R')
            ax.set_ylabel('Z')
            ax.set_aspect('equal')

            return l

        if only2D:
            ax = kwargs.pop('ax', None)
            if ax is None:
                # extra logic to handle OMFIT's auto-figure making when double-clicked in tree
                fig = pyplot.gcf()
                if len(fig.axes):
                    # normal situation in which existing figures should be respected and left alone
                    fig, ax = pyplot.subplots(subplot_kw={'aspect': 'equal'})
                else:
                    # OMFIT (or the pyplot.gcf needed to check OMFIT) made a empty figure for us
                    ax = pyplot.subplot(111, aspect='equal')
            plot_2d(ax)
        else:
            fig = kwargs.pop('figure', None)
            if fig is None:
                # extra logic to handle OMFIT's auto-figure making when double-clicked in tree
                fig = pyplot.gcf()
                if len(fig.axes):
                    # normal situation in which existing figures should be respected and left alone
                    fig = figure()
                else:
                    # OMFIT (or the pyplot.gcf needed to check OMFIT) made a empty figure for us
                    pass

            # 1D plots
            ax = pyplot.subplot(322)
            self['q'].plot(ax)
            pyplot.setp(ax.get_xticklabels(), visible=False)
            ax.set_xlabel('')
            ax.yaxis.set_major_locator(pyplot.MaxNLocator(nbins=4))
            ax = pyplot.subplot(324, sharex=ax)
            self['p'].plot(ax)
            pyplot.setp(ax.get_xticklabels(), visible=False)
            ax.set_xlabel('')
            ax.yaxis.set_major_locator(pyplot.MaxNLocator(nbins=4))
            ax = pyplot.subplot(326, sharex=ax)
            self['f'].plot(ax)
            ax.yaxis.set_major_locator(pyplot.MaxNLocator(nbins=4))

            ax = pyplot.subplot(131)
            ax.set_frame_on(False)
            ax.xaxis.set_ticks_position('bottom')
            ax.yaxis.set_ticks_position('left')
            ax.locator_params(nbins=3)
            plot_2d(ax)
